"""This file implements the robot specific pose tools."""
import math

import attr
import numpy as np

from pybullet_envs.minitaur.envs_v2.utilities import laikago_pose_utils
from pybullet_envs.minitaur.envs_v2.utilities import mini_cheetah_pose_utils
from pybullet_envs.minitaur.envs_v2.utilities import minitaur_pose_utils
from pybullet_envs.minitaur.robots import laikago
from pybullet_envs.minitaur.robots import laikago_v2
from pybullet_envs.minitaur.robots import mini_cheetah
from pybullet_envs.minitaur.robots import minitaur_v2

_ABDUCTION_ACTION_INDEXES = [0, 3, 6, 9]

# The default values used to give a neutral pose for minitaur.
_MINITAUR_DEFAULT_EXTENSION_POS = math.pi / 2
_MINITAUR_DEFAULT_SWING_POS = 0

_LAIKAGO_NEUTRAL_POSE_HIP_ANGLE = math.pi / 4
_LAIKAGO_NEUTRAL_POSE_KNEE_ANGLE = -math.pi / 2
_LAIKAGO_EXTENSION_CONVERSION_MULTIPLIER = 1.0
_LAIKAGO_SWING_CONVERSION_MULTIPLIER = -1.0

_MINI_CHEETAH_NEUTRAL_POSE_HIP_ANGLE = -math.pi / 4
_MINI_CHEETAH_NEUTRAL_POSE_KNEE_ANGLE = math.pi / 2
_MINI_CHEETAH_EXTENSION_CONVERSION_MULTIPLIER = -1.0
_MINI_CHEETAH_SWING_CONVERSION_MULTIPLIER = 1.0


def get_neutral_motor_angles(robot_class):
  """Return a neutral (standing) pose for a given robot type.

  Args:
    robot_class: This returns the class (not the instance) for the robot.
      Currently it supports minitaur, laikago and mini-cheetah.

  Returns:
    Pose object for the given robot. It's either MinitaurPose, LaikagoPose or
    MiniCheetahPose.

  Raises:
    ValueError: If the given robot_class is different than the supported robots.
  """
  if str(robot_class) in [
      str(minitaur_v2.Minitaur)
  ]:
    init_pose = minitaur_pose_utils.leg_pose_to_motor_angles(
        np.array(
            attr.astuple(
                minitaur_pose_utils.MinitaurPose(
                    swing_angle_0=_MINITAUR_DEFAULT_SWING_POS,
                    swing_angle_1=_MINITAUR_DEFAULT_SWING_POS,
                    swing_angle_2=_MINITAUR_DEFAULT_SWING_POS,
                    swing_angle_3=_MINITAUR_DEFAULT_SWING_POS,
                    extension_angle_0=_MINITAUR_DEFAULT_EXTENSION_POS,
                    extension_angle_1=_MINITAUR_DEFAULT_EXTENSION_POS,
                    extension_angle_2=_MINITAUR_DEFAULT_EXTENSION_POS,
                    extension_angle_3=_MINITAUR_DEFAULT_EXTENSION_POS))))
  elif str(robot_class) in [
      str(laikago.Laikago),
      str(laikago_v2.Laikago),
  ]:
    init_pose = np.array(
        attr.astuple(
            laikago_pose_utils.LaikagoPose(
                abduction_angle_0=0,
                hip_angle_0=_LAIKAGO_NEUTRAL_POSE_HIP_ANGLE,
                knee_angle_0=_LAIKAGO_NEUTRAL_POSE_KNEE_ANGLE,
                abduction_angle_1=0,
                hip_angle_1=_LAIKAGO_NEUTRAL_POSE_HIP_ANGLE,
                knee_angle_1=_LAIKAGO_NEUTRAL_POSE_KNEE_ANGLE,
                abduction_angle_2=0,
                hip_angle_2=_LAIKAGO_NEUTRAL_POSE_HIP_ANGLE,
                knee_angle_2=_LAIKAGO_NEUTRAL_POSE_KNEE_ANGLE,
                abduction_angle_3=0,
                hip_angle_3=_LAIKAGO_NEUTRAL_POSE_HIP_ANGLE,
                knee_angle_3=_LAIKAGO_NEUTRAL_POSE_KNEE_ANGLE)))
  elif str(robot_class) == str(mini_cheetah.MiniCheetah):
    init_pose = np.array(
        attr.astuple(
            mini_cheetah_pose_utils.MiniCheetahPose(
                abduction_angle_0=0,
                hip_angle_0=_MINI_CHEETAH_NEUTRAL_POSE_HIP_ANGLE,
                knee_angle_0=_MINI_CHEETAH_NEUTRAL_POSE_KNEE_ANGLE,
                abduction_angle_1=0,
                hip_angle_1=_MINI_CHEETAH_NEUTRAL_POSE_HIP_ANGLE,
                knee_angle_1=_MINI_CHEETAH_NEUTRAL_POSE_KNEE_ANGLE,
                abduction_angle_2=0,
                hip_angle_2=_MINI_CHEETAH_NEUTRAL_POSE_HIP_ANGLE,
                knee_angle_2=_MINI_CHEETAH_NEUTRAL_POSE_KNEE_ANGLE,
                abduction_angle_3=0,
                hip_angle_3=_MINI_CHEETAH_NEUTRAL_POSE_HIP_ANGLE,
                knee_angle_3=_MINI_CHEETAH_NEUTRAL_POSE_KNEE_ANGLE)))
  else:
    init_pose = robot_class.get_neutral_motor_angles()
  return init_pose


def convert_leg_pose_to_motor_angles(robot_class, leg_poses):
  """Convert swing-extend coordinate space to motor angles for a robot type.

  Args:
    robot_class: This returns the class (not the instance) for the robot.
      Currently it supports minitaur, laikago and mini-cheetah.
    leg_poses: A list of leg poses in [swing,extend] or [abduction, swing,
      extend] space for all 4 legs. The order is [abd_0, swing_0, extend_0,
      abd_1, swing_1, extend_1, ...] or [swing_0, extend_0, swing_1, extend_1,
      ...]. Zero swing and zero extend gives a neutral standing pose for all the
      robots. For minitaur, the conversion is fully accurate, for laikago and
      mini-cheetah the conversion is approximate where swing is reflected to hip
      and extend is reflected to both knee and the hip.

  Returns:
    List of motor positions for the selected robot. The list include 8 or 12
    motor angles depending on the given robot type as an argument. Currently
    laikago and mini-cheetah has motors for abduction which does not exist for
    minitaur robot.

  Raises:
    ValueError: Conversion fails due to wrong inputs.
  """
  default_leg_order = ["front_left", "back_left", "front_right", "back_right"]
  leg_order = default_leg_order
  if len(leg_poses) not in [8, 12]:
    raise ValueError("Dimension of the leg pose provided is not 8 or 12.")
  neutral_motor_angles = get_neutral_motor_angles(robot_class)
  motor_angles = leg_poses
  # If it is a robot with 12 motors but the provided leg pose does not contain
  # abduction, extend the pose to include abduction.
  if len(neutral_motor_angles) == 12 and len(leg_poses) == 8:
    for i in _ABDUCTION_ACTION_INDEXES:
      motor_angles.insert(i, 0)
  # If the robot does not have abduction (minitaur) but the input contains them,
  # ignore the abduction angles for the conversion.
  elif len(neutral_motor_angles) == 8 and len(leg_poses) == 12:
    del leg_poses[::3]
  # Minitaur specific conversion calculations using minitaur-specific safety
  # limits.
  if str(robot_class) in [
            
      str(minitaur_v2.Minitaur)
  ]:
    motor_angles = minitaur_pose_utils.leg_pose_to_motor_angles_with_half_pi_offset_and_safety(
        leg_poses)
  # Laikago and mini-cheetah specific conversion calculations.
  elif str(robot_class) in [
      str(mini_cheetah.MiniCheetah),
      str(laikago.Laikago),
      str(laikago_v2.Laikago),
      
  ]:
    swing_scale = 1.0
    extension_scale = 1.0
    # Laikago specific conversion multipliers.
    if str(robot_class) in [
        str(laikago.Laikago),
        str(laikago_v2.Laikago),
        
    ]:
      swing_scale = _LAIKAGO_SWING_CONVERSION_MULTIPLIER
      extension_scale = _LAIKAGO_EXTENSION_CONVERSION_MULTIPLIER
      leg_order = ["front_right", "front_left", "back_right", "back_left"]
    # Mini-cheetah specific multipliers.
    elif str(robot_class) in [str(mini_cheetah.MiniCheetah)]:
      swing_scale = _MINI_CHEETAH_SWING_CONVERSION_MULTIPLIER
      extension_scale = _MINI_CHEETAH_EXTENSION_CONVERSION_MULTIPLIER
    # In this approximate conversion for mini-cheetah and laikago we set hip
    # angle swing + half of the extend and knee angle to extend as rotation.
    # We also scale swing and extend based on some hand-tuned constants.
    multipliers = np.array([1.0, swing_scale, extension_scale] * 4)
    swing_extend_scaled = leg_poses * multipliers
    # Swing is (swing - half of the extension) due to the geometry of the leg.
    extra_swing = swing_extend_scaled * ([0, 0, -0.5] * 4)
    swing_extend_scaled += np.roll(extra_swing, -1)
    motor_angles = list(swing_extend_scaled)
    motor_angles = neutral_motor_angles + motor_angles
    # Change the order of the legs if it is different for the specific robot.
    if leg_order != default_leg_order:
      leg_order = [default_leg_order.index(leg) for leg in leg_order]
      ordered_motor_angles = []
      for i in leg_order:
        ordered_motor_angles.extend(motor_angles[3 * i:3 * i + 3])
      motor_angles = ordered_motor_angles
  else:
    motor_angles = robot_class.convert_leg_pose_to_motor_angles(leg_poses)

  return motor_angles



import os
import copy
import math

import numpy as np


class Racecar:

  def __init__(self, bullet_client, urdfRootPath='', timeStep=0.01):
    self.urdfRootPath = urdfRootPath
    self.timeStep = timeStep
    self._p = bullet_client
    self.reset()

  def reset(self):
    car = self._p.loadURDF(os.path.join(self.urdfRootPath, "racecar/racecar_differential.urdf"),
                           [0, 0, .2],
                           useFixedBase=False)
    self.racecarUniqueId = car
    #for i in range (self._p.getNumJoints(car)):
    #	print (self._p.getJointInfo(car,i))
    for wheel in range(self._p.getNumJoints(car)):
      self._p.setJointMotorControl2(car,
                                    wheel,
                                    self._p.VELOCITY_CONTROL,
                                    targetVelocity=0,
                                    force=0)
      self._p.getJointInfo(car, wheel)

    #self._p.setJointMotorControl2(car,10,self._p.VELOCITY_CONTROL,targetVelocity=1,force=10)
    c = self._p.createConstraint(car,
                                 9,
                                 car,
                                 11,
                                 jointType=self._p.JOINT_GEAR,
                                 jointAxis=[0, 1, 0],
                                 parentFramePosition=[0, 0, 0],
                                 childFramePosition=[0, 0, 0])
    self._p.changeConstraint(c, gearRatio=1, maxForce=10000)

    c = self._p.createConstraint(car,
                                 10,
                                 car,
                                 13,
                                 jointType=self._p.JOINT_GEAR,
                                 jointAxis=[0, 1, 0],
                                 parentFramePosition=[0, 0, 0],
                                 childFramePosition=[0, 0, 0])
    self._p.changeConstraint(c, gearRatio=-1, maxForce=10000)

    c = self._p.createConstraint(car,
                                 9,
                                 car,
                                 13,
                                 jointType=self._p.JOINT_GEAR,
                                 jointAxis=[0, 1, 0],
                                 parentFramePosition=[0, 0, 0],
                                 childFramePosition=[0, 0, 0])
    self._p.changeConstraint(c, gearRatio=-1, maxForce=10000)

    c = self._p.createConstraint(car,
                                 16,
                                 car,
                                 18,
                                 jointType=self._p.JOINT_GEAR,
                                 jointAxis=[0, 1, 0],
                                 parentFramePosition=[0, 0, 0],
                                 childFramePosition=[0, 0, 0])
    self._p.changeConstraint(c, gearRatio=1, maxForce=10000)

    c = self._p.createConstraint(car,
                                 16,
                                 car,
                                 19,
                                 jointType=self._p.JOINT_GEAR,
                                 jointAxis=[0, 1, 0],
                                 parentFramePosition=[0, 0, 0],
                                 childFramePosition=[0, 0, 0])
    self._p.changeConstraint(c, gearRatio=-1, maxForce=10000)

    c = self._p.createConstraint(car,
                                 17,
                                 car,
                                 19,
                                 jointType=self._p.JOINT_GEAR,
                                 jointAxis=[0, 1, 0],
                                 parentFramePosition=[0, 0, 0],
                                 childFramePosition=[0, 0, 0])
    self._p.changeConstraint(c, gearRatio=-1, maxForce=10000)

    c = self._p.createConstraint(car,
                                 1,
                                 car,
                                 18,
                                 jointType=self._p.JOINT_GEAR,
                                 jointAxis=[0, 1, 0],
                                 parentFramePosition=[0, 0, 0],
                                 childFramePosition=[0, 0, 0])
    self._p.changeConstraint(c, gearRatio=-1, gearAuxLink=15, maxForce=10000)
    c = self._p.createConstraint(car,
                                 3,
                                 car,
                                 19,
                                 jointType=self._p.JOINT_GEAR,
                                 jointAxis=[0, 1, 0],
                                 parentFramePosition=[0, 0, 0],
                                 childFramePosition=[0, 0, 0])
    self._p.changeConstraint(c, gearRatio=-1, gearAuxLink=15, maxForce=10000)

    self.steeringLinks = [0, 2]
    self.maxForce = 20
    self.nMotors = 2
    self.motorizedwheels = [8, 15]
    self.speedMultiplier = 20.
    self.steeringMultiplier = 0.5

  def getActionDimension(self):
    return self.nMotors

  def getObservationDimension(self):
    return len(self.getObservation())

  def getObservation(self):
    observation = []
    pos, orn = self._p.getBasePositionAndOrientation(self.racecarUniqueId)

    observation.extend(list(pos))
    observation.extend(list(orn))

    return observation

  def applyAction(self, motorCommands):
    targetVelocity = motorCommands[0] * self.speedMultiplier
    #print("targetVelocity")
    #print(targetVelocity)
    steeringAngle = motorCommands[1] * self.steeringMultiplier
    #print("steeringAngle")
    #print(steeringAngle)
    #print("maxForce")
    #print(self.maxForce)

    for motor in self.motorizedwheels:
      self._p.setJointMotorControl2(self.racecarUniqueId,
                                    motor,
                                    self._p.VELOCITY_CONTROL,
                                    targetVelocity=targetVelocity,
                                    force=self.maxForce)
    for steer in self.steeringLinks:
      self._p.setJointMotorControl2(self.racecarUniqueId,
                                    steer,
                                    self._p.POSITION_CONTROL,
                                    targetPosition=steeringAngle)



import pybullet as p
import pybullet_data as pd
import os

import pybullet_data

p.connect(p.DIRECT)
p.setAdditionalSearchPath(pybullet_data.getDataPath())
name_in = os.path.join(pd.getDataPath(), "duck.obj")
name_out = "duck_vhacd.obj"
name_log = "log.txt"
p.vhacd(name_in, name_out, name_log)




import time
import os
import inspect

currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)
print("parentdir=", parentdir)
import json
from pybullet_envs.deep_mimic.learning.rl_world import RLWorld
from pybullet_envs.deep_mimic.learning.ppo_agent import PPOAgent

import pybullet_data
from pybullet_utils.arg_parser import ArgParser
from pybullet_utils.logger import Logger
from pybullet_envs.deep_mimic.env.pybullet_deep_mimic_env_multiclip import PyBulletDeepMimicEnvMultiClip
import sys
import random

update_timestep = 1. / 240.
animating = True
step = False
total_reward = 0
steps = 0


def update_world(world, time_elapsed):
    timeStep = update_timestep
    world.update(timeStep)
    reward = world.env.calc_reward(agent_id=0)
    global total_reward
    total_reward += reward
    global steps
    steps += 1

    end_episode = world.env.is_episode_end()
    if (end_episode or steps >= 1000):
        print("total_reward=", total_reward)
        total_reward = 0
        steps = 0
        world.end_episode()
        world.reset()


def build_arg_parser(args):
    arg_parser = ArgParser()
    arg_parser.load_args(args)

    arg_file = arg_parser.parse_string('arg_file', '')
    if arg_file == '':
        arg_file = "run_humanoid3d_backflip_args.txt"
    if (arg_file != ''):
        path = pybullet_data.getDataPath() + "/args/" + arg_file
        if os.path.isfile(path):
            succ = arg_parser.load_file(path)
        else:
            files = [arg_parser.load_file(f) for f in os.listdir(path) if os.path.isfile(os.path.join(path, f))]
            succ = all(files)
        Logger.print2(arg_file)
        assert succ, Logger.print2('Failed to load args from: ' + arg_file)
    return arg_parser


args = sys.argv[1:]


def build_world(args, enable_draw):
    arg_parser = build_arg_parser(args)
    print("enable_draw=", enable_draw)
    env = PyBulletDeepMimicEnvMultiClip(arg_parser, enable_draw)
    world = RLWorld(env, arg_parser)

    motion_file = arg_parser.parse_string("motion_file")
    print("motion_file=", motion_file)
    bodies = arg_parser.parse_ints("fall_contact_bodies")
    print("bodies=", bodies)
    int_output_path = arg_parser.parse_string("int_output_path")
    print("int_output_path=", int_output_path)
    agent_files = pybullet_data.getDataPath() + "/" + arg_parser.parse_string("agent_files")

    AGENT_TYPE_KEY = "AgentType"

    print("agent_file=", agent_files)
    with open(agent_files) as data_file:
        json_data = json.load(data_file)
        print("json_data=", json_data)
        assert AGENT_TYPE_KEY in json_data
        agent_type = json_data[AGENT_TYPE_KEY]
        print("agent_type=", agent_type)
        agent = PPOAgent(world, id, json_data)

        agent.set_enable_training(False)
        world.reset()
    return world


if __name__ == '__main__':

    world = build_world(args, True)
    while (world.env._pybullet_client.isConnected()):

        timeStep = update_timestep
        time.sleep(timeStep)
        keys = world.env.getKeyboardEvents()

        if world.env.isKeyTriggered(keys, ' '):
            animating = not animating
        if world.env.isKeyTriggered(keys, 'i'):
            step = True
        if (animating or step):
            update_world(world, timeStep)
            step = False



# Lint as: python3
"""The inverse kinematic utilities."""
from typing import Optional, Sequence

import numpy as np

from pybullet_utils import bullet_client

_IDENTITY_ROTATION_QUAT = (0, 0, 0, 1)
_IK_SOLVER_TYPE = 0
_LINK_POS_INDEX = 0


def joint_angles_from_link_positions(
    pybullet_client: bullet_client.BulletClient,
    urdf_id: int,
    link_ids: Sequence[int],
    link_positions: Sequence[Sequence[float]],
    positions_are_in_world_frame: bool = False,
    joint_dof_ids: Optional[Sequence[int]] = None,
) -> np.ndarray:
  """Uses Inverse Kinematics to calculate joint angles.

  Args:
    pybullet_client: The bullet client.
    urdf_id: The unique id returned after loading URDF.
    link_ids: The link ids to compute the IK.
    link_positions: The (x, y, z) of the links in the body or the world frame,
      depending on whether the argument link_position_in_world_frame is true.
    positions_are_in_world_frame: Whether the input link positions are specified
      in the world frame or the robot's base frame.
    joint_dof_ids: The degrees of freedom index of the joints we want to extract
      the angles. This can be different from the joint unique ids. For example,
      a fixed joint will increase the joint unique id but will not increase the
      number of degree of freedoms. The joint dof id can be extracted in
      PyBullet by getJointInfo, which corresponds to the "qIndex" in the
      returned values. If not specified, will return all movable joint angles.

  Returns:
    A list of joint angles.
  """
  if positions_are_in_world_frame:
    world_link_positions = link_positions
  else:
    # The PyBullet inverse Kinematics Calculation depends on the current URDF
    # position/orientation, and we cannot pass them to the API. So we have to
    # always query the current base position/orientation to compute world frame
    # link positions.
    urdf_base_position, urdf_base_orientation = (
        pybullet_client.getBasePositionAndOrientation(urdf_id))
    world_link_positions = []
    for link_position in link_positions:
      world_link_position, _ = pybullet_client.multiplyTransforms(
          urdf_base_position, urdf_base_orientation, link_position,
          _IDENTITY_ROTATION_QUAT)
      world_link_positions.append(world_link_position)

  # Notice that the API expects the link positions in the world frame.
  all_joint_angles = pybullet_client.calculateInverseKinematics2(
      urdf_id, link_ids, world_link_positions, solver=_IK_SOLVER_TYPE)

  # Extract the relevant joint angles.
  if joint_dof_ids is None:
    return np.array(all_joint_angles)

  return np.array([all_joint_angles[i] for i in joint_dof_ids])


def link_position_in_world_frame(
    pybullet_client: bullet_client.BulletClient,
    urdf_id: int,
    link_id: int,
):
  """Computes the link's position in the world frame.

  Args:
    pybullet_client: The bullet client.
    urdf_id: The unique id returned after loading URDF.
    link_id: The link id to calculate its position.

  Returns:
    The position of the link in the world frame.
  """
  return np.array(pybullet_client.getLinkState(urdf_id, link_id)[0])


def link_position_in_base_frame(
    pybullet_client: bullet_client.BulletClient,
    urdf_id: int,
    link_id: int,
    base_link_id: Optional[int] = None,
):
  """Computes the link's local position in the robot frame.

  Args:
    pybullet_client: The bullet client.
    urdf_id: The unique id returned after loading URDF.
    link_id: The link to calculate its relative position.
    base_link_id: The link id of the base. For the kinematics robot, such as
      wheeled_robot_base, three additional joints are added to connect the world
      and the base. The base_link_id is no longer -1, and need to be passed in.


  Returns:
    The relative position of the link.
  """
  if base_link_id is None:
    base_position, base_orientation = (
        pybullet_client.getBasePositionAndOrientation(urdf_id))
  else:
    base_link_state = pybullet_client.getLinkState(urdf_id, base_link_id)
    base_position, base_orientation = base_link_state[0], base_link_state[1]

  inverse_translation, inverse_rotation = pybullet_client.invertTransform(
      base_position, base_orientation)

  link_state = pybullet_client.getLinkState(urdf_id, link_id)
  link_position = link_state[0]
  link_local_position, _ = pybullet_client.multiplyTransforms(
      inverse_translation, inverse_rotation, link_position, (0, 0, 0, 1))

  return np.array(link_local_position)


def compute_jacobian(
    pybullet_client: bullet_client.BulletClient,
    urdf_id: int,
    link_id: int,
    all_joint_positions: Sequence[float],
    additional_translation: Optional[Sequence[float]] = (0, 0, 0),
) -> np.ndarray:
  """Computes the Jacobian matrix for the given point on a link.

  CAVEAT: If during URDF loading process additional rotations are provided, the
  computed Jacobian are also transformed.

  Args:
    pybullet_client: The bullet client.
    urdf_id: The unique id returned after loading URDF.
    link_id: The link id as returned from loadURDF.
    all_joint_positions: all the joint positions of the robot. This should
      include the dummy/kinematic drive joints for the wheeled robot.
    additional_translation: The additional translation of the point in the link
      CoM frame.

  Returns:
    The 3 x N transposed Jacobian matrix. where N is the total DoFs of the
    robot. For a quadruped, the first 6 columns of the matrix corresponds to
    the CoM translation and rotation. The columns corresponds to a leg can be
    extracted with indices [6 + leg_id * 3: 6 + leg_id * 3 + 3].
  """

  zero_vec = [0] * len(all_joint_positions)
  jv, _ = pybullet_client.calculateJacobian(
      urdf_id,
      link_id,
      additional_translation,
      all_joint_positions,
      objVelocities=zero_vec,
      objAccelerations=zero_vec)
  jacobian = np.array(jv)
  assert jacobian.shape[0] == 3
  return jacobian


def rotate_to_base_frame(
    pybullet_client: bullet_client.BulletClient,
    urdf_id: int,
    vector: Sequence[float],
    init_orientation_inv_quat: Optional[Sequence[float]] = (0, 0, 0, 1)
) -> np.ndarray:
  """Rotates the input vector to the base coordinate systems.

  Note: This is different from world frame to base frame transformation, as we
  do not apply any translation here.

  Args:
    pybullet_client: The bullet client.
    urdf_id: The unique id returned after loading URDF.
    vector: Input vector in the world frame.
    init_orientation_inv_quat:

  Returns:
    A rotated vector in the base frame.
  """
  _, base_orientation_quat = (
      pybullet_client.getBasePositionAndOrientation(urdf_id))
  _, base_orientation_quat_from_init = pybullet_client.multiplyTransforms(
      positionA=(0, 0, 0),
      orientationA=init_orientation_inv_quat,
      positionB=(0, 0, 0),
      orientationB=base_orientation_quat)
  _, inverse_base_orientation = pybullet_client.invertTransform(
      [0, 0, 0], base_orientation_quat_from_init)

  # PyBullet transforms requires simple list/tuple or it may crash.
  if isinstance(vector, np.ndarray):
    vector_list = vector.tolist()
  else:
    vector_list = vector

  local_vector, _ = pybullet_client.multiplyTransforms(
      positionA=(0, 0, 0),
      orientationA=inverse_base_orientation,
      positionB=vector_list,
      orientationB=(0, 0, 0, 1),
  )
  return np.array(local_vector)



import os
import inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)
import pybullet_envs
import gym
import argparse
import pybullet as p


def test(args):
  count = 0
  env = gym.make(args.env)
  env.env.configure(args)
  
  print("args.render=", args.render)
  if (args.render == 1):
    env.render(mode="human")
  env.reset()
  
  w, h, vmat,projmat,camup,camfwd,hor,ver,yaw,pitch,dist,target= p.getDebugVisualizerCamera()
  dist = 0.4
  yaw = 0
  p.resetDebugVisualizerCamera(dist,yaw, pitch,target)
  
  for obindex in range (p.getNumBodies()):
    obuid = p.getBodyUniqueId(obindex)
    p.changeDynamics(obuid, -1, linearDamping=0, angularDamping=0)
    for l in range (p.getNumJoints(obuid)):
      p.changeDynamics(obuid, l, linearDamping=0, angularDamping=0, jointDamping=0)
      #if (l==0):
      #  p.setJointMotorControl2(obuid,l,p.POSITION_CONTROL,targetPosition=0)
      if (l==2):
        jp,jv,_,_ = p.getJointState(obuid,l)
        p.resetJointState(obuid,l, jp,0.01 )
    
  if (args.resetbenchmark):
    while (1):
      env.reset()
      print("p.getNumBodies()=", p.getNumBodies())
      print("count=", count)
      count += 1
  print("action space:")
  sample = env.action_space.sample()
  action = sample * 0.0
  action = [0,0]#sample * 0.0
  
  print("action=")
  print(action)
  for i in range(args.steps):
    obs, rewards, done, _ = env.step(action)
    if (args.rgb):
      print(env.render(mode="rgb_array"))
    print("obs=")
    print(obs)
    print("rewards")
    print(rewards)
    print("done")
    print(done)


def main():
  import argparse
  parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
  parser.add_argument('--env', help='environment ID', default='AntBulletEnv-v0')
  parser.add_argument('--seed', help='RNG seed', type=int, default=0)
  parser.add_argument('--render', help='OpenGL Visualizer', type=int, default=0)
  parser.add_argument('--rgb', help='rgb_array gym rendering', type=int, default=0)
  parser.add_argument('--resetbenchmark',
                      help='Repeat reset to show reset performance',
                      type=int,
                      default=0)
  parser.add_argument('--steps', help='Number of steps', type=int, default=1)

  args = parser.parse_args()
  test(args)


if __name__ == '__main__':
  main()



import numpy as np
import copy as copy
try:
  import tensorflow.compat.v1 as tf
except Exception:
  import tensorflow as tf

from pybullet_envs.deep_mimic.learning.pg_agent import PGAgent
from pybullet_envs.deep_mimic.learning.solvers.mpi_solver import MPISolver
import pybullet_envs.deep_mimic.learning.tf_util as TFUtil
import pybullet_envs.deep_mimic.learning.rl_util as RLUtil
from pybullet_utils.logger import Logger
import pybullet_utils.mpi_util as MPIUtil
import pybullet_utils.math_util as MathUtil
from pybullet_envs.deep_mimic.env.env import Env
'''
Proximal Policy Optimization Agent
'''


class PPOAgent(PGAgent):
  NAME = "PPO"
  EPOCHS_KEY = "Epochs"
  BATCH_SIZE_KEY = "BatchSize"
  RATIO_CLIP_KEY = "RatioClip"
  NORM_ADV_CLIP_KEY = "NormAdvClip"
  TD_LAMBDA_KEY = "TDLambda"
  TAR_CLIP_FRAC = "TarClipFrac"
  ACTOR_STEPSIZE_DECAY = "ActorStepsizeDecay"

  def __init__(self, world, id, json_data):
    super().__init__(world, id, json_data)
    return

  def _load_params(self, json_data):
    super()._load_params(json_data)

    self.epochs = 1 if (self.EPOCHS_KEY not in json_data) else json_data[self.EPOCHS_KEY]
    self.batch_size = 1024 if (
        self.BATCH_SIZE_KEY not in json_data) else json_data[self.BATCH_SIZE_KEY]
    self.ratio_clip = 0.2 if (
        self.RATIO_CLIP_KEY not in json_data) else json_data[self.RATIO_CLIP_KEY]
    self.norm_adv_clip = 5 if (
        self.NORM_ADV_CLIP_KEY not in json_data) else json_data[self.NORM_ADV_CLIP_KEY]
    self.td_lambda = 0.95 if (
        self.TD_LAMBDA_KEY not in json_data) else json_data[self.TD_LAMBDA_KEY]
    self.tar_clip_frac = -1 if (
        self.TAR_CLIP_FRAC not in json_data) else json_data[self.TAR_CLIP_FRAC]
    self.actor_stepsize_decay = 0.5 if (
        self.ACTOR_STEPSIZE_DECAY not in json_data) else json_data[self.ACTOR_STEPSIZE_DECAY]

    num_procs = MPIUtil.get_num_procs()
    local_batch_size = int(self.batch_size / num_procs)
    min_replay_size = 2 * local_batch_size  # needed to prevent buffer overflow
    assert (self.replay_buffer_size > min_replay_size)

    self.replay_buffer_size = np.maximum(min_replay_size, self.replay_buffer_size)

    return

  def _build_nets(self, json_data):
    assert self.ACTOR_NET_KEY in json_data
    assert self.CRITIC_NET_KEY in json_data

    actor_net_name = json_data[self.ACTOR_NET_KEY]
    critic_net_name = json_data[self.CRITIC_NET_KEY]
    actor_init_output_scale = 1 if (self.ACTOR_INIT_OUTPUT_SCALE_KEY not in json_data
                                   ) else json_data[self.ACTOR_INIT_OUTPUT_SCALE_KEY]

    s_size = self.get_state_size()
    g_size = self.get_goal_size()
    a_size = self.get_action_size()

    # setup input tensors
    self.s_tf = tf.placeholder(tf.float32, shape=[None, s_size], name="s")
    self.a_tf = tf.placeholder(tf.float32, shape=[None, a_size], name="a")
    self.tar_val_tf = tf.placeholder(tf.float32, shape=[None], name="tar_val")
    self.adv_tf = tf.placeholder(tf.float32, shape=[None], name="adv")
    self.g_tf = tf.placeholder(tf.float32,
                               shape=([None, g_size] if self.has_goal() else None),
                               name="g")
    self.old_logp_tf = tf.placeholder(tf.float32, shape=[None], name="old_logp")
    self.exp_mask_tf = tf.placeholder(tf.float32, shape=[None], name="exp_mask")

    with tf.variable_scope('main'):
      with tf.variable_scope('actor'):
        self.a_mean_tf = self._build_net_actor(actor_net_name, actor_init_output_scale)
      with tf.variable_scope('critic'):
        self.critic_tf = self._build_net_critic(critic_net_name)

    if (self.a_mean_tf != None):
      Logger.print2('Built actor net: ' + actor_net_name)

    if (self.critic_tf != None):
      Logger.print2('Built critic net: ' + critic_net_name)

    self.norm_a_std_tf = self.exp_params_curr.noise * tf.ones(a_size)
    norm_a_noise_tf = self.norm_a_std_tf * tf.random_normal(shape=tf.shape(self.a_mean_tf))
    norm_a_noise_tf *= tf.expand_dims(self.exp_mask_tf, axis=-1)
    self.sample_a_tf = self.a_mean_tf + norm_a_noise_tf * self.a_norm.std_tf
    self.sample_a_logp_tf = TFUtil.calc_logp_gaussian(x_tf=norm_a_noise_tf,
                                                      mean_tf=None,
                                                      std_tf=self.norm_a_std_tf)

    return

  def _build_losses(self, json_data):
    actor_weight_decay = 0 if (
        self.ACTOR_WEIGHT_DECAY_KEY not in json_data) else json_data[self.ACTOR_WEIGHT_DECAY_KEY]
    critic_weight_decay = 0 if (
        self.CRITIC_WEIGHT_DECAY_KEY not in json_data) else json_data[self.CRITIC_WEIGHT_DECAY_KEY]

    norm_val_diff = self.val_norm.normalize_tf(self.tar_val_tf) - self.val_norm.normalize_tf(
        self.critic_tf)
    self.critic_loss_tf = 0.5 * tf.reduce_mean(tf.square(norm_val_diff))

    if (critic_weight_decay != 0):
      self.critic_loss_tf += critic_weight_decay * self._weight_decay_loss('main/critic')

    norm_tar_a_tf = self.a_norm.normalize_tf(self.a_tf)
    self._norm_a_mean_tf = self.a_norm.normalize_tf(self.a_mean_tf)

    self.logp_tf = TFUtil.calc_logp_gaussian(norm_tar_a_tf, self._norm_a_mean_tf,
                                             self.norm_a_std_tf)
    ratio_tf = tf.exp(self.logp_tf - self.old_logp_tf)
    actor_loss0 = self.adv_tf * ratio_tf
    actor_loss1 = self.adv_tf * tf.clip_by_value(ratio_tf, 1.0 - self.ratio_clip,
                                                 1 + self.ratio_clip)
    self.actor_loss_tf = -tf.reduce_mean(tf.minimum(actor_loss0, actor_loss1))

    norm_a_bound_min = self.a_norm.normalize(self.a_bound_min)
    norm_a_bound_max = self.a_norm.normalize(self.a_bound_max)
    a_bound_loss = TFUtil.calc_bound_loss(self._norm_a_mean_tf, norm_a_bound_min, norm_a_bound_max)
    self.actor_loss_tf += a_bound_loss

    if (actor_weight_decay != 0):
      self.actor_loss_tf += actor_weight_decay * self._weight_decay_loss('main/actor')

    # for debugging
    self.clip_frac_tf = tf.reduce_mean(
        tf.to_float(tf.greater(tf.abs(ratio_tf - 1.0), self.ratio_clip)))

    return

  def _build_solvers(self, json_data):
    actor_stepsize = 0.001 if (
        self.ACTOR_STEPSIZE_KEY not in json_data) else json_data[self.ACTOR_STEPSIZE_KEY]
    actor_momentum = 0.9 if (
        self.ACTOR_MOMENTUM_KEY not in json_data) else json_data[self.ACTOR_MOMENTUM_KEY]
    critic_stepsize = 0.01 if (
        self.CRITIC_STEPSIZE_KEY not in json_data) else json_data[self.CRITIC_STEPSIZE_KEY]
    critic_momentum = 0.9 if (
        self.CRITIC_MOMENTUM_KEY not in json_data) else json_data[self.CRITIC_MOMENTUM_KEY]

    critic_vars = self._tf_vars('main/critic')
    critic_opt = tf.train.MomentumOptimizer(learning_rate=critic_stepsize,
                                            momentum=critic_momentum)
    self.critic_grad_tf = tf.gradients(self.critic_loss_tf, critic_vars)
    self.critic_solver = MPISolver(self.sess, critic_opt, critic_vars)

    self._actor_stepsize_tf = tf.get_variable(dtype=tf.float32,
                                              name='actor_stepsize',
                                              initializer=actor_stepsize,
                                              trainable=False)
    self._actor_stepsize_ph = tf.get_variable(dtype=tf.float32, name='actor_stepsize_ph', shape=[])
    self._actor_stepsize_update_op = self._actor_stepsize_tf.assign(self._actor_stepsize_ph)

    actor_vars = self._tf_vars('main/actor')
    actor_opt = tf.train.MomentumOptimizer(learning_rate=self._actor_stepsize_tf,
                                           momentum=actor_momentum)
    self.actor_grad_tf = tf.gradients(self.actor_loss_tf, actor_vars)
    self.actor_solver = MPISolver(self.sess, actor_opt, actor_vars)

    return

  def _decide_action(self, s, g):
    with self.sess.as_default(), self.graph.as_default():
      self._exp_action = self._enable_stoch_policy() and MathUtil.flip_coin(
          self.exp_params_curr.rate)
      #print("_decide_action._exp_action=",self._exp_action)
      a, logp = self._eval_actor(s, g, self._exp_action)
    return a[0], logp[0]

  def _eval_actor(self, s, g, enable_exp):
    s = np.reshape(s, [-1, self.get_state_size()])
    g = np.reshape(g, [-1, self.get_goal_size()]) if self.has_goal() else None

    feed = {self.s_tf: s, self.g_tf: g, self.exp_mask_tf: np.array([1 if enable_exp else 0])}

    a, logp = self.sess.run([self.sample_a_tf, self.sample_a_logp_tf], feed_dict=feed)
    return a, logp

  def _train_step(self):
    adv_eps = 1e-5

    start_idx = self.replay_buffer.buffer_tail
    end_idx = self.replay_buffer.buffer_head
    assert (start_idx == 0)
    assert (self.replay_buffer.get_current_size() <= self.replay_buffer.buffer_size
           )  # must avoid overflow
    assert (start_idx < end_idx)

    idx = np.array(list(range(start_idx, end_idx)))
    end_mask = self.replay_buffer.is_path_end(idx)
    end_mask = np.logical_not(end_mask)

    vals = self._compute_batch_vals(start_idx, end_idx)
    new_vals = self._compute_batch_new_vals(start_idx, end_idx, vals)

    valid_idx = idx[end_mask]
    exp_idx = self.replay_buffer.get_idx_filtered(self.EXP_ACTION_FLAG).copy()
    num_valid_idx = valid_idx.shape[0]
    num_exp_idx = exp_idx.shape[0]
    exp_idx = np.column_stack([exp_idx, np.array(list(range(0, num_exp_idx)), dtype=np.int32)])

    local_sample_count = valid_idx.size
    global_sample_count = int(MPIUtil.reduce_sum(local_sample_count))
    mini_batches = int(np.ceil(global_sample_count / self.mini_batch_size))

    adv = new_vals[exp_idx[:, 0]] - vals[exp_idx[:, 0]]
    new_vals = np.clip(new_vals, self.val_min, self.val_max)

    adv_mean = np.mean(adv)
    adv_std = np.std(adv)
    adv = (adv - adv_mean) / (adv_std + adv_eps)
    adv = np.clip(adv, -self.norm_adv_clip, self.norm_adv_clip)

    critic_loss = 0
    actor_loss = 0
    actor_clip_frac = 0

    for e in range(self.epochs):
      np.random.shuffle(valid_idx)
      np.random.shuffle(exp_idx)

      for b in range(mini_batches):
        batch_idx_beg = b * self._local_mini_batch_size
        batch_idx_end = batch_idx_beg + self._local_mini_batch_size

        critic_batch = np.array(range(batch_idx_beg, batch_idx_end), dtype=np.int32)
        actor_batch = critic_batch.copy()
        critic_batch = np.mod(critic_batch, num_valid_idx)
        actor_batch = np.mod(actor_batch, num_exp_idx)
        shuffle_actor = (actor_batch[-1] < actor_batch[0]) or (actor_batch[-1] == num_exp_idx - 1)

        critic_batch = valid_idx[critic_batch]
        actor_batch = exp_idx[actor_batch]
        critic_batch_vals = new_vals[critic_batch]
        actor_batch_adv = adv[actor_batch[:, 1]]

        critic_s = self.replay_buffer.get('states', critic_batch)
        critic_g = self.replay_buffer.get('goals', critic_batch) if self.has_goal() else None
        curr_critic_loss = self._update_critic(critic_s, critic_g, critic_batch_vals)

        actor_s = self.replay_buffer.get("states", actor_batch[:, 0])
        actor_g = self.replay_buffer.get("goals", actor_batch[:, 0]) if self.has_goal() else None
        actor_a = self.replay_buffer.get("actions", actor_batch[:, 0])
        actor_logp = self.replay_buffer.get("logps", actor_batch[:, 0])
        curr_actor_loss, curr_actor_clip_frac = self._update_actor(actor_s, actor_g, actor_a,
                                                                   actor_logp, actor_batch_adv)

        critic_loss += curr_critic_loss
        actor_loss += np.abs(curr_actor_loss)
        actor_clip_frac += curr_actor_clip_frac

        if (shuffle_actor):
          np.random.shuffle(exp_idx)

    total_batches = mini_batches * self.epochs
    critic_loss /= total_batches
    actor_loss /= total_batches
    actor_clip_frac /= total_batches

    critic_loss = MPIUtil.reduce_avg(critic_loss)
    actor_loss = MPIUtil.reduce_avg(actor_loss)
    actor_clip_frac = MPIUtil.reduce_avg(actor_clip_frac)

    critic_stepsize = self.critic_solver.get_stepsize()
    actor_stepsize = self.update_actor_stepsize(actor_clip_frac)

    self.logger.log_tabular('Critic_Loss', critic_loss)
    self.logger.log_tabular('Critic_Stepsize', critic_stepsize)
    self.logger.log_tabular('Actor_Loss', actor_loss)
    self.logger.log_tabular('Actor_Stepsize', actor_stepsize)
    self.logger.log_tabular('Clip_Frac', actor_clip_frac)
    self.logger.log_tabular('Adv_Mean', adv_mean)
    self.logger.log_tabular('Adv_Std', adv_std)

    self.replay_buffer.clear()

    return

  def _get_iters_per_update(self):
    return 1

  def _valid_train_step(self):
    samples = self.replay_buffer.get_current_size()
    exp_samples = self.replay_buffer.count_filtered(self.EXP_ACTION_FLAG)
    global_sample_count = int(MPIUtil.reduce_sum(samples))
    global_exp_min = int(MPIUtil.reduce_min(exp_samples))
    return (global_sample_count > self.batch_size) and (global_exp_min > 0)

  def _compute_batch_vals(self, start_idx, end_idx):
    states = self.replay_buffer.get_all("states")[start_idx:end_idx]
    goals = self.replay_buffer.get_all("goals")[start_idx:end_idx] if self.has_goal() else None

    idx = np.array(list(range(start_idx, end_idx)))
    is_end = self.replay_buffer.is_path_end(idx)
    is_fail = self.replay_buffer.check_terminal_flag(idx, Env.Terminate.Fail)
    is_succ = self.replay_buffer.check_terminal_flag(idx, Env.Terminate.Succ)
    is_fail = np.logical_and(is_end, is_fail)
    is_succ = np.logical_and(is_end, is_succ)

    vals = self._eval_critic(states, goals)
    vals[is_fail] = self.val_fail
    vals[is_succ] = self.val_succ

    return vals

  def _compute_batch_new_vals(self, start_idx, end_idx, val_buffer):
    rewards = self.replay_buffer.get_all("rewards")[start_idx:end_idx]

    if self.discount == 0:
      new_vals = rewards.copy()
    else:
      new_vals = np.zeros_like(val_buffer)

      curr_idx = start_idx
      while curr_idx < end_idx:
        idx0 = curr_idx - start_idx
        idx1 = self.replay_buffer.get_path_end(curr_idx) - start_idx
        r = rewards[idx0:idx1]
        v = val_buffer[idx0:(idx1 + 1)]

        new_vals[idx0:idx1] = RLUtil.compute_return(r, self.discount, self.td_lambda, v)
        curr_idx = idx1 + start_idx + 1

    return new_vals

  def _update_critic(self, s, g, tar_vals):
    feed = {self.s_tf: s, self.g_tf: g, self.tar_val_tf: tar_vals}

    loss, grads = self.sess.run([self.critic_loss_tf, self.critic_grad_tf], feed)
    self.critic_solver.update(grads)
    return loss

  def _update_actor(self, s, g, a, logp, adv):
    feed = {self.s_tf: s, self.g_tf: g, self.a_tf: a, self.adv_tf: adv, self.old_logp_tf: logp}

    loss, grads, clip_frac = self.sess.run(
        [self.actor_loss_tf, self.actor_grad_tf, self.clip_frac_tf], feed)
    self.actor_solver.update(grads)

    return loss, clip_frac

  def update_actor_stepsize(self, clip_frac):
    clip_tol = 1.5
    step_scale = 2
    max_stepsize = 1e-2
    min_stepsize = 1e-8
    warmup_iters = 5

    actor_stepsize = self.actor_solver.get_stepsize()
    if (self.tar_clip_frac >= 0 and self.iter > warmup_iters):
      min_clip = self.tar_clip_frac / clip_tol
      max_clip = self.tar_clip_frac * clip_tol
      under_tol = clip_frac < min_clip
      over_tol = clip_frac > max_clip

      if (over_tol or under_tol):
        if (over_tol):
          actor_stepsize *= self.actor_stepsize_decay
        else:
          actor_stepsize /= self.actor_stepsize_decay

        actor_stepsize = np.clip(actor_stepsize, min_stepsize, max_stepsize)
        self.set_actor_stepsize(actor_stepsize)

    return actor_stepsize

  def set_actor_stepsize(self, stepsize):
    feed = {
        self._actor_stepsize_ph: stepsize,
    }
    self.sess.run(self._actor_stepsize_update_op, feed)
    return



r"""ScenarioSet example for Laikago MPC controller.

blaze run -c opt \
//robotics/reinforcement_learning/minitaur/agents/baseline_controller\
:locomotion_controller_in_scenario_set_example -- --gait=slow_trot \
--add_random_push=True
"""

from absl import app
from absl import flags
import gin
import numpy as np
import scipy.interpolate

from pybullet_envs.minitaur.agents.baseline_controller import locomotion_controller_setup
from pybullet_envs.minitaur.envs_v2 import env_loader

FLAGS = flags.FLAGS

SCENARIO_SET_CONFIG = """
import pybullet_envs.minitaur.envs_v2.scenarios.locomotion_simple_scenario_set

include "google3/robotics/reinforcement_learning/minitaur/envs_v2/scenarios/default_scenario_set.gin"

default_scenario_set/singleton.constructor = @locomotion_simple_scenario_set.LocomotionSimpleScenarioSet


locomotion_simple_scenario_set.LocomotionSimpleScenarioSet.selector = "flat_ground"
locomotion_gym_env.LocomotionGymEnv.task = @scenario_set.task()
locomotion_gym_env.LocomotionGymEnv.scene = @scenario_set.scene()
locomotion_gym_env.LocomotionGymEnv.env_randomizers = [
  @scenario_set.env_randomizer()
]
"""

_MAX_TIME_SECONDS = 30

flags.DEFINE_enum("gait", "fast_trot",
                  ["fast_trot", "slow_trot", "walk", "stand"],
                  "The gait pattern to use")

flags.DEFINE_boolean("add_random_push", False,
                     "whether to add random push to the robot in simulation")


def _start_stop_profile(max_speed=0.5, axis=0, duration=3):
  speed_profile = np.zeros((3, 4))

  speed_profile[1, axis] = max_speed

  return (0, 0.5, duration + 0.5), speed_profile.tolist()


def _random_speed_profile(max_speed=1, axis=0, time_interval=1.0):
  num_pts = 11
  time_points = np.arange(num_pts) * time_interval

  speed_profile = np.zeros((num_pts, 4))
  speed_profile[:, axis] = np.random.uniform(0, max_speed, num_pts)
  speed_profile[-1, :] = 0
  return time_points.tolist(), speed_profile.tolist()


def _body_height_profile(z_range=(0.3, 0.55)):
  del z_range
  # TODO(tingnan): Implement this.


def _generate_linear_angular_speed(t, time_points, speed_points):
  """Creates an example speed profile based on time for demo purpose."""

  speed = scipy.interpolate.interp1d(
      time_points,
      speed_points,
      kind="previous",
      fill_value="extrapolate",
      axis=0)(
          t)

  return speed[0:3], speed[3]


def _update_controller_params(controller, lin_speed, ang_speed):
  controller.swing_leg_controller.desired_speed = lin_speed
  controller.swing_leg_controller.desired_twisting_speed = ang_speed
  controller.stance_leg_controller.desired_speed = lin_speed
  controller.stance_leg_controller.desired_twisting_speed = ang_speed


def _gen_stability_test_start_stop():
  """Generates the speed profile for start/stop tests."""
  axis_to_name = {
      0: "velocity x",
      1: "velocity y",
      3: "angular velocity z",
  }

  axis_to_max_speed = {
      0: 1.0,
      1: 0.5,
      3: 2.5,
  }

  gait_multiplier = {
      "slow_trot": 0.7,
      "walk": 0.3,
      "fast_trot": 1.0,
  }

  for axis in (0, 1, 3):
    yield axis_to_name[axis], _start_stop_profile(
        axis_to_max_speed[axis] * gait_multiplier[FLAGS.gait], axis)


def _gen_stability_test_random():
  """Generates the speed profile for random walking tests."""
  axis_to_name = {
      0: "velocity x",
      1: "velocity y",
      3: "angular velocity z",
  }

  axis_to_max_speed = {
      0: 1.0,
      1: 0.5,
      3: 2.5,
  }

  gait_multiplier = {
      "slow_trot": 0.7,
      "walk": 0.3,
      "fast_trot": 1.0,
  }

  for axis in (0, 1, 3):
    yield axis_to_name[axis], _random_speed_profile(
        axis_to_max_speed[axis] * gait_multiplier[FLAGS.gait], axis)


def _test_stability(max_time=5, render=False, test_generator=None):
  """Tests the stability of the controller using speed profiles."""
  locomotion_controller_setup.load_sim_config(render=render)
  gin.parse_config(SCENARIO_SET_CONFIG)
  if FLAGS.add_random_push:
    locomotion_controller_setup.add_random_push_config()

  env = env_loader.load()
  controller = locomotion_controller_setup.setup_controller(
      env.robot, gait=FLAGS.gait)

  for name, speed_profile in test_generator():
    env.reset()
    controller.reset()
    current_time = 0
    while current_time < max_time:
      current_time = env.get_time_since_reset()
      lin_speed, ang_speed = _generate_linear_angular_speed(
          current_time, speed_profile[0], speed_profile[1])
      _update_controller_params(controller, lin_speed, ang_speed)

      # Needed before every call to get_action().
      controller.update()
      hybrid_action = controller.get_action()

      _, _, done, _ = env.step(hybrid_action)
      if done:
        break
    print(f"Scene name: flat ground. Random push: {FLAGS.add_random_push}. "
          f"Survival time for {name} = {speed_profile[1]} is {current_time}")


def main(argv):
  del argv
  _test_stability(render=True, test_generator=_gen_stability_test_start_stop)
  _test_stability(
      max_time=15, render=True, test_generator=_gen_stability_test_random)


if __name__ == "__main__":
  app.run(main)



# Copyright 2017 The TensorFlow Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Wrap a dictionary to access keys as attributes."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import contextlib


class AttrDict(dict):
  """Wrap a dictionary to access keys as attributes."""

  def __init__(self, *args, **kwargs):
    super(AttrDict, self).__init__(*args, **kwargs)
    super(AttrDict, self).__setattr__('_mutable', False)

  def __getattr__(self, key):
    # Do not provide None for unimplemented magic attributes.
    if key.startswith('__'):
      raise AttributeError
    return self.get(key, None)

  def __setattr__(self, key, value):
    if not self._mutable:
      message = "Cannot set attribute '{}'.".format(key)
      message += " Use 'with obj.unlocked:' scope to set attributes."
      raise RuntimeError(message)
    if key.startswith('__'):
      raise AttributeError("Cannot set magic attribute '{}'".format(key))
    self[key] = value

  @property
  @contextlib.contextmanager
  def unlocked(self):
    super(AttrDict, self).__setattr__('_mutable', True)
    yield
    super(AttrDict, self).__setattr__('_mutable', False)

  def copy(self):
    return type(self)(super(AttrDict, self).copy())



import json
import math


class MotionCaptureData(object):

  def __init__(self):
    self.Reset()

  def Reset(self):
    self._motion_data = []

  def Load(self, path):
    with open(path, 'r') as f:
      self._motion_data = json.load(f)

  def NumFrames(self):
    return len(self._motion_data['Frames'])

  def KeyFrameDuraction(self):
    return self._motion_data['Frames'][0][0]

  def getCycleTime(self):
    keyFrameDuration = self.KeyFrameDuraction()
    cycleTime = keyFrameDuration * (self.NumFrames() - 1)
    return cycleTime

  def calcCycleCount(self, simTime, cycleTime):
    phases = simTime / cycleTime
    count = math.floor(phases)
    loop = True
    #count = (loop) ? count : cMathUtil::Clamp(count, 0, 1);
    return count

  def computeCycleOffset(self):
    firstFrame = 0
    lastFrame = self.NumFrames() - 1
    frameData = self._motion_data['Frames'][0]
    frameDataNext = self._motion_data['Frames'][lastFrame]

    basePosStart = [frameData[1], frameData[2], frameData[3]]
    basePosEnd = [frameDataNext[1], frameDataNext[2], frameDataNext[3]]
    self._cycleOffset = [
        basePosEnd[0] - basePosStart[0], basePosEnd[1] - basePosStart[1],
        basePosEnd[2] - basePosStart[2]
    ]
    return self._cycleOffset



"""Common tools and functionalities used in different tasks."""
import numpy as np


def calculate_target_speed_at_timestep(speed_stages):
  """Interpolates the speed based on a speed profile and simulation steps.

  Args:
    speed_stages: The list of timesteps (in increasing order) and speed (float
      or a list of floats for x and y components) at that specific timestep.
      Example formats can be found on task_utils_test.py and at the header of
      the speed_reward_task.py.

  Returns:
    Target speed for the specific step (meters per simulation step).
  Raises:
    ValueError if the input is not in the expected format.
  """
  if len(speed_stages) != 2 or len(speed_stages[0]) != len(speed_stages[1]):
    raise ValueError('Speed stages for the task is not in correct format!')
  steps = np.array(speed_stages[0])
  speeds = np.array(speed_stages[1])
  num_steps = steps[-1]
  if len(speeds.shape) == 1:
    return np.interp(range(num_steps), steps, speeds)
  speed_at_timestep = np.interp(range(num_steps), steps, speeds[:, 0]).reshape(
      (num_steps, 1))
  speed_at_timestep = speed_at_timestep.reshape((num_steps, 1))
  if speeds.shape[1] == 2:
    speed_y = np.interp(range(num_steps), steps, speeds[:, 1]).reshape(
        (num_steps, 1))
    speed_at_timestep = np.concatenate((speed_at_timestep, speed_y), axis=1)
  else:
    speed_at_timestep = np.concatenate(
        (speed_at_timestep, np.zeros((num_steps, 1))), axis=1)
  return speed_at_timestep


def calculate_distance(vector_1, vector_2):
  """Calculates the distance between 2 vectors.

  This is used to calculate distance between 2 points in 3D space as well as
  distances between two orientation vectors.

  Args:
    vector_1: First vector that will be used for comparison with the other.
    vector_2: Second vector used for comparison.

  Returns:
    Distance between the two vectors represented by a float.
  """
  return np.linalg.norm(np.array(vector_1) - np.array(vector_2))


def turn_angle(new_vector, reference_vector):
  """Calculates the change in orientation of the two vectors.

    This is used to calculate the relative angle perception for the
    robot.

  Args:
    new_vector: The front vector of the robot at current timestep.
    reference_vector: The front vector of the robot at previous timestep.

  Returns:
    Angle representing the change in orientation of the robot projected to
    x-y plane.
  """
  # Project the vectors to x-y plane
  v1 = np.resize(new_vector, 3)
  v2 = np.resize(reference_vector, 3)
  v1[2] = 0
  v2[2] = 0
  # Calculate the right hand rotation between two vectors using z vector
  # (0,0,-1) as reference cross product vector.
  # Compatible with the yaw rotation of pyBullet.
  return np.arctan2(np.dot(np.cross(v2, v1), (0, 0, -1)), np.dot(v1, v2))


def front_vector(pybullet_client, orientation):
  """Calculates the front vector of the robot on x-y plane.

  Args:
    pybullet_client: Pybullet client instantiation.
    orientation: Orientation of the robot in quaternion form.

  Returns:
    3D vector where z component is set to 0.
  """
  rot_matrix = pybullet_client.getMatrixFromQuaternion(orientation)
  return [rot_matrix[0], -rot_matrix[1], 0]


def calculate_estimated_energy_consumption(motor_torques, motor_velocities,
                                           sim_time_step, num_action_repeat):
  """Calculates energy consumption based on the args listed.

  Args:
    motor_torques: Torques of all the motors
    motor_velocities: Velocities of all the motors.
    sim_time_step: Simulation time step length (seconds).
    num_action_repeat: How many steps the simulation repeats the same action.

  Returns:
    Total energy consumption of all the motors (watts).
  """
  return np.abs(np.dot(motor_torques,
                       motor_velocities)) * sim_time_step * num_action_repeat



import pybullet as p
from time import sleep
import matplotlib.pyplot as plt
import numpy as np
import pybullet_data

physicsClient = p.connect(p.GUI)

p.setAdditionalSearchPath(pybullet_data.getDataPath())
p.setGravity(0, 0, 0)
bearStartPos1 = [-3.3, 0, 0]
bearStartOrientation1 = p.getQuaternionFromEuler([0, 0, 0])
bearId1 = p.loadURDF("plane.urdf", bearStartPos1, bearStartOrientation1)
bearStartPos2 = [0, 0, 0]
bearStartOrientation2 = p.getQuaternionFromEuler([0, 0, 0])
bearId2 = p.loadURDF("teddy_large.urdf", bearStartPos2, bearStartOrientation2)
textureId = p.loadTexture("checker_grid.jpg")
#p.changeVisualShape(objectUniqueId=0, linkIndex=-1, textureUniqueId=textureId)
#p.changeVisualShape(objectUniqueId=1, linkIndex=-1, textureUniqueId=textureId)

useRealTimeSimulation = 1

if (useRealTimeSimulation):
  p.setRealTimeSimulation(1)

while 1:
  if (useRealTimeSimulation):
    camera = p.getDebugVisualizerCamera()
    viewMat = camera[2]
    projMat = camera[3]
    #An example of setting the view matrix for the projective texture.
    #viewMat = p.computeViewMatrix(cameraEyePosition=[7,0,0], cameraTargetPosition=[0,0,0], cameraUpVector=[0,0,1])
    p.getCameraImage(300,
                     300,
                     renderer=p.ER_BULLET_HARDWARE_OPENGL,
                     flags=p.ER_USE_PROJECTIVE_TEXTURE,
                     projectiveTextureView=viewMat,
                     projectiveTextureProj=projMat)
    p.setGravity(0, 0, 0)
  else:
    p.stepSimulation()



"""A simple locomotion taskand termination condition."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

import gin
from pybullet_envs.minitaur.envs_v2.tasks import task_interface
from pybullet_envs.minitaur.envs_v2.tasks import task_utils
from pybullet_envs.minitaur.envs_v2.tasks import terminal_conditions
from pybullet_envs.minitaur.envs_v2.utilities import env_utils_v2 as env_utils


@gin.configurable
class SimpleForwardTask(task_interface.Task):
  """A basic "move forward" task."""

  def __init__(self,
               weight=1.0,
               terminal_condition=terminal_conditions
               .default_terminal_condition_for_minitaur,
               divide_with_dt=False,
               clip_velocity=None,
               energy_penalty_coef=0.0,
               min_com_height=None,
               weight_action_accel=None):
    """Initializes the task.

    Args:
      weight: Float. The scaling factor for the reward.
      terminal_condition: Callable object or function. Determines if the task is
        done.
      divide_with_dt: if True, we divide the velocity reward with dt.
      clip_velocity: if not None, we will clip the velocity with this value.
      energy_penalty_coef: Coefficient for the energy penalty that will be added
        to the reward. 0 by default.
      min_com_height: Minimum height for the center of mass of the robot that
        will be used to terminate the task. This is used to obtain task specific
        gaits and set by the config or gin files based on the task and robot.
      weight_action_accel: if not None, penalize the action acceleration.

    Raises:
      ValueError: The energey coefficient is smaller than zero.
    """
    self._weight = weight
    self._terminal_condition = terminal_condition
    self._last_base_position = None
    self._divide_with_dt = divide_with_dt
    self._clip_velocity = clip_velocity
    self._weight_action_accel = weight_action_accel
    self._action_history_sensor = None
    self._min_com_height = min_com_height
    self._energy_penalty_coef = energy_penalty_coef
    self._env = None
    self._step_count = 0
    if energy_penalty_coef < 0:
      raise ValueError("Energy Penalty Coefficient should be >= 0")

  def __call__(self, env):
    return self.reward(env)

  def reset(self, env):
    self._env = env
    self._last_base_position = env_utils.get_robot_base_position(
        self._env.robot)

    if self._weight_action_accel is not None:
      sensor_name = "LastAction"
      self._action_history_sensor = env.sensor_by_name(sensor_name)

  @property
  def step_count(self):
    return self._step_count

  def update(self, env):
    """Updates the internal state of the task."""
    del env
    self._last_base_position = env_utils.get_robot_base_position(
        self._env.robot)

  def reward(self, env):
    """Get the reward without side effects."""
    del env

    self._step_count += 1
    env = self._env
    current_base_position = env_utils.get_robot_base_position(self._env.robot)
    velocity = current_base_position[0] - self._last_base_position[0]
    if self._divide_with_dt:
      velocity /= env.env_time_step
    if self._clip_velocity is not None:
      limit = float(self._clip_velocity)
      velocity = np.clip(velocity, -limit, limit)

    if self._weight_action_accel is None:
      action_acceleration_penalty = 0.0
    else:
      past_actions = self._action_history_sensor.get_observation().T
      action = past_actions[0]
      prev_action = past_actions[1]
      prev_prev_action = past_actions[2]
      acc = action - 2 * prev_action + prev_prev_action
      action_acceleration_penalty = (
          float(self._weight_action_accel) * np.mean(np.abs(acc)))

    reward = velocity
    reward -= action_acceleration_penalty

    # Energy
    if self._energy_penalty_coef > 0:
      energy_reward = -task_utils.calculate_estimated_energy_consumption(
          self._env.robot.motor_torques, self._env.robot.motor_velocities,
          self._env.sim_time_step, self._env.num_action_repeat)
      reward += energy_reward * self._energy_penalty_coef

    return reward * self._weight

  def done(self, env):
    del env
    position = env_utils.get_robot_base_position(self._env.robot)
    if self._min_com_height and position[2] < self._min_com_height:
      return True
    return self._terminal_condition(self._env)



import pybullet as p
import time
import math
from datetime import datetime
import pybullet_data

clid = p.connect(p.SHARED_MEMORY)
if (clid < 0):
  p.connect(p.GUI)

p.setAdditionalSearchPath(pybullet_data.getDataPath())
p.loadURDF("plane.urdf", [0, 0, -1.3])
p.configureDebugVisualizer(p.COV_ENABLE_RENDERING, 0)
sawyerId = p.loadURDF("pole.urdf", [0, 0, 0])
p.configureDebugVisualizer(p.COV_ENABLE_RENDERING, 1)
p.resetBasePositionAndOrientation(sawyerId, [0, 0, 0], [0, 0, 0, 1])

sawyerEndEffectorIndex = 3
numJoints = p.getNumJoints(sawyerId)
#joint damping coefficents
jd = [0.1, 0.1, 0.1, 0.1]

p.setGravity(0, 0, 0)
t = 0.
prevPose = [0, 0, 0]
prevPose1 = [0, 0, 0]
hasPrevPose = 0

ikSolver = 0
useRealTimeSimulation = 0
p.setRealTimeSimulation(useRealTimeSimulation)
#trailDuration is duration (in seconds) after debug lines will be removed automatically
#use 0 for no-removal
trailDuration = 1

while 1:
  if (useRealTimeSimulation):
    dt = datetime.now()
    t = (dt.second / 60.) * 2. * math.pi
  else:
    t = t + 0.01
    time.sleep(0.01)

  for i in range(1):
    pos = [2. * math.cos(t), 2. * math.cos(t), 0. + 2. * math.sin(t)]
    jointPoses = p.calculateInverseKinematics(sawyerId,
                                              sawyerEndEffectorIndex,
                                              pos,
                                              jointDamping=jd,
                                              solver=ikSolver,
                                              maxNumIterations=100)

    #reset the joint state (ignoring all dynamics, not recommended to use during simulation)
    for i in range(numJoints):
      jointInfo = p.getJointInfo(sawyerId, i)
      qIndex = jointInfo[3]
      if qIndex > -1:
        p.resetJointState(sawyerId, i, jointPoses[qIndex - 7])

  ls = p.getLinkState(sawyerId, sawyerEndEffectorIndex)
  if (hasPrevPose):
    p.addUserDebugLine(prevPose, pos, [0, 0, 0.3], 1, trailDuration)
    p.addUserDebugLine(prevPose1, ls[4], [1, 0, 0], 1, trailDuration)
  prevPose = pos
  prevPose1 = ls[4]
  hasPrevPose = 1



"""Abstract base class for environment randomizer."""

import abc


class EnvRandomizerBase(object):
  """Abstract base class for environment randomizer.

  Randomizes physical parameters of the objects in the simulation and adds
  perturbations to the stepping of the simulation.
  """

  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def randomize_env(self, env):
    """Randomize the simulated_objects in the environment.

    Will be called at when env is reset. The physical parameters will be fixed
    for that episode and be randomized again in the next environment.reset().

    Args:
      env: The Minitaur gym environment to be randomized.
    """
    pass

  def randomize_step(self, env):
    """Randomize simulation steps.

    Will be called at every timestep. May add random forces/torques to Minitaur.

    Args:
      env: The Minitaur gym environment to be randomized.
    """
    pass



"""This file implements the functionalities of a minitaur derpy using pybullet.

It is the result of first pass system identification for the derpy robot. The


"""
import math

import os, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)

import numpy as np
from pybullet_envs.minitaur.envs import minitaur

KNEE_CONSTRAINT_POINT_LONG = [0, 0.0055, 0.088]
KNEE_CONSTRAINT_POINT_SHORT = [0, 0.0055, 0.100]


class MinitaurDerpy(minitaur.Minitaur):
  """The minitaur class that simulates a quadruped robot from Ghost Robotics.

  """

  def Reset(self, reload_urdf=True, default_motor_angles=None, reset_time=3.0):
    """Reset the minitaur to its initial states.

    Args:
      reload_urdf: Whether to reload the urdf file. If not, Reset() just place
        the minitaur back to its starting position.
      default_motor_angles: The default motor angles. If it is None, minitaur
        will hold a default pose (motor angle math.pi / 2) for 100 steps. In
        torque control mode, the phase of holding the default pose is skipped.
      reset_time: The duration (in seconds) to hold the default motor angles. If
        reset_time <= 0 or in torque control mode, the phase of holding the
        default pose is skipped.
    """
    if self._on_rack:
      init_position = minitaur.INIT_RACK_POSITION
    else:
      init_position = minitaur.INIT_POSITION
    if reload_urdf:
      if self._self_collision_enabled:
        self.quadruped = self._pybullet_client.loadURDF(
            "%s/quadruped/minitaur_derpy.urdf" % self._urdf_root,
            init_position,
            useFixedBase=self._on_rack,
            flags=(self._pybullet_client.URDF_USE_SELF_COLLISION_EXCLUDE_PARENT))
      else:
        self.quadruped = self._pybullet_client.loadURDF("%s/quadruped/minitaur_derpy.urdf" %
                                                        self._urdf_root,
                                                        init_position,
                                                        useFixedBase=self._on_rack)
      self._BuildJointNameToIdDict()
      self._BuildUrdfIds()
      if self._remove_default_joint_damping:
        self._RemoveDefaultJointDamping()
      self._BuildMotorIdList()
      self._RecordMassInfoFromURDF()
      self._RecordInertiaInfoFromURDF()
      self.ResetPose(add_constraint=True)
    else:
      self._pybullet_client.resetBasePositionAndOrientation(self.quadruped, init_position,
                                                            minitaur.INIT_ORIENTATION)
      self._pybullet_client.resetBaseVelocity(self.quadruped, [0, 0, 0], [0, 0, 0])
      self.ResetPose(add_constraint=False)

    self._overheat_counter = np.zeros(self.num_motors)
    self._motor_enabled_list = [True] * self.num_motors
    self._step_counter = 0

    # Perform reset motion within reset_duration if in position control mode.
    # Nothing is performed if in torque control mode for now.
    # TODO(jietan): Add reset motion when the torque control is fully supported.
    self._observation_history.clear()
    if not self._torque_control_enabled and reset_time > 0.0:
      self.ReceiveObservation()
      for _ in range(100):
        self.ApplyAction([math.pi / 2] * self.num_motors)
        self._pybullet_client.stepSimulation()
        self.ReceiveObservation()
      if default_motor_angles is not None:
        num_steps_to_reset = int(reset_time / self.time_step)
        for _ in range(num_steps_to_reset):
          self.ApplyAction(default_motor_angles)
          self._pybullet_client.stepSimulation()
          self.ReceiveObservation()
    self.ReceiveObservation()

  def _ResetPoseForLeg(self, leg_id, add_constraint):
    """Reset the initial pose for the leg.

    Args:
      leg_id: It should be 0, 1, 2, or 3, which represents the leg at
        front_left, back_left, front_right and back_right.
      add_constraint: Whether to add a constraint at the joints of two feet.
    """
    knee_friction_force = 0
    half_pi = math.pi / 2.0
    knee_angle = -2.1834

    leg_position = minitaur.LEG_POSITION[leg_id]
    self._pybullet_client.resetJointState(self.quadruped,
                                          self._joint_name_to_id["motor_" + leg_position +
                                                                 "L_joint"],
                                          self._motor_direction[2 * leg_id] * half_pi,
                                          targetVelocity=0)
    self._pybullet_client.resetJointState(self.quadruped,
                                          self._joint_name_to_id["knee_" + leg_position +
                                                                 "L_joint"],
                                          self._motor_direction[2 * leg_id] * knee_angle,
                                          targetVelocity=0)
    self._pybullet_client.resetJointState(self.quadruped,
                                          self._joint_name_to_id["motor_" + leg_position +
                                                                 "R_joint"],
                                          self._motor_direction[2 * leg_id + 1] * half_pi,
                                          targetVelocity=0)
    self._pybullet_client.resetJointState(self.quadruped,
                                          self._joint_name_to_id["knee_" + leg_position +
                                                                 "R_joint"],
                                          self._motor_direction[2 * leg_id + 1] * knee_angle,
                                          targetVelocity=0)
    if add_constraint:
      if leg_id < 2:
        self._pybullet_client.createConstraint(
            self.quadruped, self._joint_name_to_id["knee_" + leg_position + "R_joint"],
            self.quadruped, self._joint_name_to_id["knee_" + leg_position + "L_joint"],
            self._pybullet_client.JOINT_POINT2POINT, [0, 0, 0], KNEE_CONSTRAINT_POINT_SHORT,
            KNEE_CONSTRAINT_POINT_LONG)
      else:
        self._pybullet_client.createConstraint(
            self.quadruped, self._joint_name_to_id["knee_" + leg_position + "R_joint"],
            self.quadruped, self._joint_name_to_id["knee_" + leg_position + "L_joint"],
            self._pybullet_client.JOINT_POINT2POINT, [0, 0, 0], KNEE_CONSTRAINT_POINT_LONG,
            KNEE_CONSTRAINT_POINT_SHORT)

    if self._accurate_motor_model_enabled or self._pd_control_enabled:
      # Disable the default motor in pybullet.
      self._pybullet_client.setJointMotorControl2(
          bodyIndex=self.quadruped,
          jointIndex=(self._joint_name_to_id["motor_" + leg_position + "L_joint"]),
          controlMode=self._pybullet_client.VELOCITY_CONTROL,
          targetVelocity=0,
          force=knee_friction_force)
      self._pybullet_client.setJointMotorControl2(
          bodyIndex=self.quadruped,
          jointIndex=(self._joint_name_to_id["motor_" + leg_position + "R_joint"]),
          controlMode=self._pybullet_client.VELOCITY_CONTROL,
          targetVelocity=0,
          force=knee_friction_force)

    else:
      self._SetDesiredMotorAngleByName("motor_" + leg_position + "L_joint",
                                       self._motor_direction[2 * leg_id] * half_pi)
      self._SetDesiredMotorAngleByName("motor_" + leg_position + "R_joint",
                                       self._motor_direction[2 * leg_id + 1] * half_pi)

    self._pybullet_client.setJointMotorControl2(
        bodyIndex=self.quadruped,
        jointIndex=(self._joint_name_to_id["knee_" + leg_position + "L_joint"]),
        controlMode=self._pybullet_client.VELOCITY_CONTROL,
        targetVelocity=0,
        force=knee_friction_force)
    self._pybullet_client.setJointMotorControl2(
        bodyIndex=self.quadruped,
        jointIndex=(self._joint_name_to_id["knee_" + leg_position + "R_joint"]),
        controlMode=self._pybullet_client.VELOCITY_CONTROL,
        targetVelocity=0,
        force=knee_friction_force)



import pybullet as p1
from pybullet_utils import bullet_client
import pybullet_data
from pybullet_utils import pd_controller_stable

import time
import motion_capture_data
import quadrupedPoseInterpolator

useKinematic = False
useConstraints = False

p = bullet_client.BulletClient(connection_mode=p1.GUI)
p.setAdditionalSearchPath(pybullet_data.getDataPath())

plane = p.loadURDF("plane.urdf")
p.setGravity(0, 0, -10)
timeStep = 1. / 500
p.setTimeStep(timeStep)
#p.setDefaultContactERP(0)
#urdfFlags = p.URDF_USE_SELF_COLLISION+p.URDF_USE_SELF_COLLISION_EXCLUDE_ALL_PARENTS
urdfFlags = p.URDF_USE_SELF_COLLISION

startPos = [0.007058990464444105, 0.03149299192130908, 0.4918981912395484]
startOrn = [0.005934649695708604, 0.7065453990917289, 0.7076373820553712, -0.0027774940359030264]
quadruped = p.loadURDF("laikago/laikago.urdf",
                       startPos,
                       startOrn,
                       flags=urdfFlags,
                       useFixedBase=False)
p.resetBasePositionAndOrientation(quadruped, startPos, startOrn)
if not useConstraints:
  for j in range(p.getNumJoints(quadruped)):
    p.setJointMotorControl2(quadruped, j, p.POSITION_CONTROL, force=0)

#This cube is added as a soft constraint to keep the laikago from falling
#since we didn't train it yet, it doesn't balance
cube = p.loadURDF("cube_no_rotation.urdf", [0, 0, -0.5], [0, 0.5, 0.5, 0])
p.setCollisionFilterGroupMask(cube, -1, 0, 0)
for j in range(p.getNumJoints(cube)):
  p.setJointMotorControl2(cube, j, p.POSITION_CONTROL, force=0)
  p.setCollisionFilterGroupMask(cube, j, 0, 0)
  p.changeVisualShape(cube, j, rgbaColor=[1, 0, 0, 0])
cid = p.createConstraint(cube,
                         p.getNumJoints(cube) - 1, quadruped, -1, p.JOINT_FIXED, [0, 0, 0],
                         [0, 1, 0], [0, 0, 0])
p.changeConstraint(cid, maxForce=10)

jointIds = []
paramIds = []
jointOffsets = []
jointDirections = [-1, 1, 1, 1, 1, 1, -1, 1, 1, 1, 1, 1]
jointAngles = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

for i in range(4):
  jointOffsets.append(0)
  jointOffsets.append(-0.7)
  jointOffsets.append(0.7)

maxForceId = p.addUserDebugParameter("maxForce", 0, 100, 120)

for j in range(p.getNumJoints(quadruped)):
  p.changeDynamics(quadruped, j, linearDamping=0, angularDamping=0)
  info = p.getJointInfo(quadruped, j)
  #print(info)
  jointName = info[1]
  jointType = info[2]
  if (jointType == p.JOINT_PRISMATIC or jointType == p.JOINT_REVOLUTE):
    jointIds.append(j)

startQ = [
    0.08389, 0.8482, -1.547832, -0.068933, 0.625726, -1.272086, 0.074398, 0.61135, -1.255892,
    -0.068262, 0.836745, -1.534517
]
for j in range(p.getNumJoints(quadruped)):
  p.resetJointState(quadruped, jointIds[j], jointDirections[j] * startQ[j] + jointOffsets[j])

qpi = quadrupedPoseInterpolator.QuadrupedPoseInterpolator()

#enable collision between lower legs

for j in range(p.getNumJoints(quadruped)):
  print(p.getJointInfo(quadruped, j))

#2,5,8 and 11 are the lower legs
lower_legs = [2, 5, 8, 11]
for l0 in lower_legs:
  for l1 in lower_legs:
    if (l1 > l0):
      enableCollision = 1
      print("collision for pair", l0, l1,
            p.getJointInfo(quadruped, l0)[12],
            p.getJointInfo(quadruped, l1)[12], "enabled=", enableCollision)
      p.setCollisionFilterPair(quadruped, quadruped, 2, 5, enableCollision)

jointIds = []
paramIds = []
jointOffsets = []
jointDirections = [-1, 1, 1, 1, 1, 1, -1, 1, 1, 1, 1, 1]
jointAngles = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

for i in range(4):
  jointOffsets.append(0)
  jointOffsets.append(-0.7)
  jointOffsets.append(0.7)

maxUpForceId = p.addUserDebugParameter("maxUpForce", 0, 100, 20)

for j in range(p.getNumJoints(quadruped)):
  p.changeDynamics(quadruped, j, linearDamping=0, angularDamping=0)
  info = p.getJointInfo(quadruped, j)
  #print(info)
  jointName = info[1]
  jointType = info[2]
  if (jointType == p.JOINT_PRISMATIC or jointType == p.JOINT_REVOLUTE):
    jointIds.append(j)

p.getCameraImage(480, 320)
p.setRealTimeSimulation(0)

joints = []

mocapData = motion_capture_data.MotionCaptureData()

motionPath = pybullet_data.getDataPath() + "/data/motions/laikago_walk.txt"

mocapData.Load(motionPath)
print("mocapData.NumFrames=", mocapData.NumFrames())
print("mocapData.KeyFrameDuraction=", mocapData.KeyFrameDuraction())
print("mocapData.getCycleTime=", mocapData.getCycleTime())
print("mocapData.computeCycleOffset=", mocapData.computeCycleOffset())

stablePD = pd_controller_stable.PDControllerStable(p)

cycleTime = mocapData.getCycleTime()
t = 0

while t < 10. * cycleTime:
  #get interpolated joint
  keyFrameDuration = mocapData.KeyFrameDuraction()
  cycleTime = mocapData.getCycleTime()
  cycleCount = mocapData.calcCycleCount(t, cycleTime)

  #print("cycleTime=",cycleTime)
  #print("cycleCount=",cycleCount)

  #print("cycles=",cycles)
  frameTime = t - cycleCount * cycleTime
  #print("frameTime=",frameTime)
  if (frameTime < 0):
    frameTime += cycleTime

  frame = int(frameTime / keyFrameDuration)
  frameNext = frame + 1
  if (frameNext >= mocapData.NumFrames()):
    frameNext = frame
  frameFraction = (frameTime - frame * keyFrameDuration) / (keyFrameDuration)
  #print("frame=",frame)
  #print("frameFraction=",frameFraction)
  frameData = mocapData._motion_data['Frames'][frame]
  frameDataNext = mocapData._motion_data['Frames'][frameNext]

  jointsStr, qdot = qpi.Slerp(frameFraction, frameData, frameDataNext, p)

  maxForce = p.readUserDebugParameter(maxForceId)
  #print("jointIds=", jointIds)
  
  maxUpForce = p.readUserDebugParameter(maxUpForceId)
  p.changeConstraint(cid, maxForce=maxUpForce)


  if useKinematic:
    basePos = startPos
    basePos =  [float(-jointsStr[0]),-float(jointsStr[1]),float(jointsStr[2])]
    baseOrn = [float(jointsStr[4]),float(jointsStr[5]),float(jointsStr[6]), float(jointsStr[3])]
    p.resetBasePositionAndOrientation(quadruped, basePos,baseOrn)
    
    for j in range(12):
        #skip the base positional dofs
        targetPos = float(jointsStr[j + 7])
        p.resetJointState(quadruped,jointIds[j],jointDirections[j] * targetPos + jointOffsets[j])
  else:
	  if useConstraints:
	    for j in range(12):
	      #skip the base positional dofs
	      targetPos = float(jointsStr[j + 7])
	      p.setJointMotorControl2(quadruped,
	                              jointIds[j],
	                              p.POSITION_CONTROL,
	                              jointDirections[j] * targetPos + jointOffsets[j],
	                              force=maxForce)
	
	  else:
	    desiredPositions = []
	    for j in range(7):
	      targetPosUnmodified = float(jointsStr[j])
	      desiredPositions.append(targetPosUnmodified)
	    for j in range(12):
	      targetPosUnmodified = float(jointsStr[j + 7])
	      targetPos = jointDirections[j] * targetPosUnmodified + jointOffsets[j]
	      desiredPositions.append(targetPos)
	    numBaseDofs = 6
	    totalDofs = 12 + numBaseDofs
	    desiredVelocities = None
	    if desiredVelocities == None:
	      desiredVelocities = [0] * totalDofs
	    taus = stablePD.computePD(bodyUniqueId=quadruped,
	                              jointIndices=jointIds,
	                              desiredPositions=desiredPositions,
	                              desiredVelocities=desiredVelocities,
	                              kps=[4000] * totalDofs,
	                              kds=[40] * totalDofs,
	                              maxForces=[maxForce] * totalDofs,
	                              timeStep=timeStep)
	
	    dofIndex = 6
	    scaling = 1
	    for index in range(len(jointIds)):
	      jointIndex = jointIds[index]
	      force = [scaling * taus[dofIndex]]
	      #print("force[", jointIndex, "]=", force)
	      p.setJointMotorControlMultiDof(quadruped,
	                                     jointIndex,
	                                     controlMode=p.TORQUE_CONTROL,
	                                     force=force)
	      dofIndex += 1
	
	  p.stepSimulation()
  t += timeStep
  time.sleep(timeStep)

useOrgData = False
if useOrgData:
  with open("data1.txt", "r") as filestream:
    for line in filestream:
      maxForce = p.readUserDebugParameter(maxForceId)
      currentline = line.split(",")
      frame = currentline[0]
      t = currentline[1]
      joints = currentline[2:14]
      for j in range(12):
        targetPos = float(joints[j])
        p.setJointMotorControl2(quadruped,
                                jointIds[j],
                                p.POSITION_CONTROL,
                                jointDirections[j] * targetPos + jointOffsets[j],
                                force=maxForce)
      p.stepSimulation()
      for lower_leg in lower_legs:
        pass
        #print("points for ", quadruped, " link: ", lower_leg)
        #pts = p.getContactPoints(quadruped, -1, lower_leg)
        #print("num points=",len(pts))
        #for pt in pts:
        #	print(pt[9])
      time.sleep(1. / 500.)

for j in range(p.getNumJoints(quadruped)):
  p.changeDynamics(quadruped, j, linearDamping=0, angularDamping=0)
  info = p.getJointInfo(quadruped, j)
  js = p.getJointState(quadruped, j)
  #print(info)
  jointName = info[1]
  jointType = info[2]
  if (jointType == p.JOINT_PRISMATIC or jointType == p.JOINT_REVOLUTE):
    paramIds.append(
        p.addUserDebugParameter(jointName.decode("utf-8"), -4, 4,
                                (js[0] - jointOffsets[j]) / jointDirections[j]))

p.setRealTimeSimulation(1)

while (1):

  for i in range(len(paramIds)):
    c = paramIds[i]
    targetPos = p.readUserDebugParameter(c)
    maxForce = p.readUserDebugParameter(maxForceId)
    p.setJointMotorControl2(quadruped,
                            jointIds[i],
                            p.POSITION_CONTROL,
                            jointDirections[i] * targetPos + jointOffsets[i],
                            force=maxForce)



from robot_bases import MJCFBasedRobot
import numpy as np


class Reacher(MJCFBasedRobot):
  TARG_LIMIT = 0.27

  def __init__(self):
    MJCFBasedRobot.__init__(self, 'reacher.xml', 'body0', action_dim=2, obs_dim=9)

  def robot_specific_reset(self, bullet_client):
    self.jdict["target_x"].reset_current_position(
        self.np_random.uniform(low=-self.TARG_LIMIT, high=self.TARG_LIMIT), 0)
    self.jdict["target_y"].reset_current_position(
        self.np_random.uniform(low=-self.TARG_LIMIT, high=self.TARG_LIMIT), 0)
    self.fingertip = self.parts["fingertip"]
    self.target = self.parts["target"]
    self.central_joint = self.jdict["joint0"]
    self.elbow_joint = self.jdict["joint1"]
    self.central_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14), 0)
    self.elbow_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14), 0)

  def apply_action(self, a):
    assert (np.isfinite(a).all())
    self.central_joint.set_motor_torque(0.05 * float(np.clip(a[0], -1, +1)))
    self.elbow_joint.set_motor_torque(0.05 * float(np.clip(a[1], -1, +1)))

  def calc_state(self):
    theta, self.theta_dot = self.central_joint.current_relative_position()
    self.gamma, self.gamma_dot = self.elbow_joint.current_relative_position()
    target_x, _ = self.jdict["target_x"].current_position()
    target_y, _ = self.jdict["target_y"].current_position()
    self.to_target_vec = np.array(self.fingertip.pose().xyz()) - np.array(self.target.pose().xyz())
    return np.array([
        target_x,
        target_y,
        self.to_target_vec[0],
        self.to_target_vec[1],
        np.cos(theta),
        np.sin(theta),
        self.theta_dot,
        self.gamma,
        self.gamma_dot,
    ])

  def calc_potential(self):
    return -100 * np.linalg.norm(self.to_target_vec)


class Pusher(MJCFBasedRobot):
  min_target_placement_radius = 0.5
  max_target_placement_radius = 0.8
  min_object_to_target_distance = 0.1
  max_object_to_target_distance = 0.4

  def __init__(self):
    MJCFBasedRobot.__init__(self, 'pusher.xml', 'body0', action_dim=7, obs_dim=55)

  def robot_specific_reset(self, bullet_client):
    # parts
    self.fingertip = self.parts["fingertip"]
    self.target = self.parts["target"]
    self.object = self.parts["object"]

    # joints
    self.shoulder_pan_joint = self.jdict["shoulder_pan_joint"]
    self.shoulder_lift_joint = self.jdict["shoulder_lift_joint"]
    self.upper_arm_roll_joint = self.jdict["upper_arm_roll_joint"]
    self.elbow_flex_joint = self.jdict["elbow_flex_joint"]
    self.forearm_roll_joint = self.jdict["forearm_roll_joint"]
    self.wrist_flex_joint = self.jdict["wrist_flex_joint"]
    self.wrist_roll_joint = self.jdict["wrist_roll_joint"]

    self.target_pos = np.concatenate([
        self.np_random.uniform(low=-1, high=1, size=1),
        self.np_random.uniform(low=-1, high=1, size=1)
    ])

    # make length of vector between min and max_target_placement_radius
    self.target_pos = self.target_pos \
          / np.linalg.norm(self.target_pos) \
          * self.np_random.uniform(low=self.min_target_placement_radius,
                 high=self.max_target_placement_radius, size=1)

    self.object_pos = np.concatenate([
        self.np_random.uniform(low=-1, high=1, size=1),
        self.np_random.uniform(low=-1, high=1, size=1)
    ])

    # make length of vector between min and max_object_to_target_distance
    self.object_pos = self.object_pos \
          / np.linalg.norm(self.object_pos - self.target_pos) \
          * self.np_random.uniform(low=self.min_object_to_target_distance,
                 high=self.max_object_to_target_distance, size=1)

    # set position of objects
    self.zero_offset = np.array([0.45, 0.55])
    self.jdict["target_x"].reset_current_position(self.target_pos[0] - self.zero_offset[0], 0)
    self.jdict["target_y"].reset_current_position(self.target_pos[1] - self.zero_offset[1], 0)
    self.jdict["object_x"].reset_current_position(self.object_pos[0] - self.zero_offset[0], 0)
    self.jdict["object_y"].reset_current_position(self.object_pos[1] - self.zero_offset[1], 0)

    # randomize all joints TODO: Will this work or do we have to constrain this resetting in some way?
    self.shoulder_pan_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14), 0)
    self.shoulder_lift_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14),
                                                    0)
    self.upper_arm_roll_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14),
                                                     0)
    self.elbow_flex_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14), 0)
    self.forearm_roll_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14), 0)
    self.wrist_flex_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14), 0)
    self.wrist_roll_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14), 0)

  def apply_action(self, a):
    assert (np.isfinite(a).all())
    self.shoulder_pan_joint.set_motor_torque(0.05 * float(np.clip(a[0], -1, +1)))
    self.shoulder_lift_joint.set_motor_torque(0.05 * float(np.clip(a[1], -1, +1)))
    self.upper_arm_roll_joint.set_motor_torque(0.05 * float(np.clip(a[2], -1, +1)))
    self.elbow_flex_joint.set_motor_torque(0.05 * float(np.clip(a[3], -1, +1)))
    self.forearm_roll_joint.set_motor_torque(0.05 * float(np.clip(a[4], -1, +1)))
    self.wrist_flex_joint.set_motor_torque(0.05 * float(np.clip(a[5], -1, +1)))
    self.wrist_roll_joint.set_motor_torque(0.05 * float(np.clip(a[6], -1, +1)))

  def calc_state(self):
    self.to_target_vec = self.target_pos - self.object_pos
    return np.concatenate([
        np.array([j.current_position() for j in self.ordered_joints]).flatten(),  # all positions
        np.array([j.current_relative_position() for j in self.ordered_joints
                 ]).flatten(),  # all speeds
        self.to_target_vec,
        self.fingertip.pose().xyz(),
        self.object.pose().xyz(),
        self.target.pose().xyz(),
    ])


class Striker(MJCFBasedRobot):
  min_target_placement_radius = 0.1
  max_target_placement_radius = 0.8
  min_object_placement_radius = 0.1
  max_object_placement_radius = 0.8

  def __init__(self):
    MJCFBasedRobot.__init__(self, 'striker.xml', 'body0', action_dim=7, obs_dim=55)

  def robot_specific_reset(self, bullet_client):
    # parts
    self.fingertip = self.parts["fingertip"]
    self.target = self.parts["target"]
    self.object = self.parts["object"]

    # joints
    self.shoulder_pan_joint = self.jdict["shoulder_pan_joint"]
    self.shoulder_lift_joint = self.jdict["shoulder_lift_joint"]
    self.upper_arm_roll_joint = self.jdict["upper_arm_roll_joint"]
    self.elbow_flex_joint = self.jdict["elbow_flex_joint"]
    self.forearm_roll_joint = self.jdict["forearm_roll_joint"]
    self.wrist_flex_joint = self.jdict["wrist_flex_joint"]
    self.wrist_roll_joint = self.jdict["wrist_roll_joint"]

    self._min_strike_dist = np.inf
    self._striked = False
    self._strike_pos = None

    # reset position and speed of manipulator
    # TODO: Will this work or do we have to constrain this resetting in some way?
    self.shoulder_pan_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14), 0)
    self.shoulder_lift_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14),
                                                    0)
    self.upper_arm_roll_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14),
                                                     0)
    self.elbow_flex_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14), 0)
    self.forearm_roll_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14), 0)
    self.wrist_flex_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14), 0)
    self.wrist_roll_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14), 0)

    self.zero_offset = np.array([0.45, 0.55, 0])
    self.object_pos = np.concatenate([
        self.np_random.uniform(low=-1, high=1, size=1),
        self.np_random.uniform(low=-1, high=1, size=1)
    ])

    # make length of vector between min and max_object_placement_radius
    self.object_pos = self.object_pos \
          / np.linalg.norm(self.object_pos) \
          * self.np_random.uniform(low=self.min_object_placement_radius,
                 high=self.max_object_placement_radius, size=1)

    # reset object position
    self.jdict["object_x"].reset_current_position(self.object_pos[0] - self.zero_offset[0], 0)
    self.jdict["object_y"].reset_current_position(self.object_pos[1] - self.zero_offset[1], 0)

    self.target_pos = np.concatenate([
        self.np_random.uniform(low=-1, high=1, size=1),
        self.np_random.uniform(low=-1, high=1, size=1),
        self.np_random.uniform(low=-1, high=1, size=1)
    ])

    # make length of vector between min and max_target_placement_radius
    self.target_pos = self.target_pos \
          / np.linalg.norm(self.target_pos) \
          * self.np_random.uniform(low=self.min_target_placement_radius,
                 high=self.max_target_placement_radius, size=1)

    self.target.reset_pose(self.target_pos - self.zero_offset, np.array([0, 0, 0, 1]))

  def apply_action(self, a):
    assert (np.isfinite(a).all())
    self.shoulder_pan_joint.set_motor_torque(0.05 * float(np.clip(a[0], -1, +1)))
    self.shoulder_lift_joint.set_motor_torque(0.05 * float(np.clip(a[1], -1, +1)))
    self.upper_arm_roll_joint.set_motor_torque(0.05 * float(np.clip(a[2], -1, +1)))
    self.elbow_flex_joint.set_motor_torque(0.05 * float(np.clip(a[3], -1, +1)))
    self.forearm_roll_joint.set_motor_torque(0.05 * float(np.clip(a[4], -1, +1)))
    self.wrist_flex_joint.set_motor_torque(0.05 * float(np.clip(a[5], -1, +1)))
    self.wrist_roll_joint.set_motor_torque(0.05 * float(np.clip(a[6], -1, +1)))

  def calc_state(self):
    self.to_target_vec = self.target_pos - self.object_pos
    return np.concatenate([
        np.array([j.current_position() for j in self.ordered_joints]).flatten(),  # all positions
        np.array([j.current_relative_position() for j in self.ordered_joints
                 ]).flatten(),  # all speeds
        self.to_target_vec,
        self.fingertip.pose().xyz(),
        self.object.pose().xyz(),
        self.target.pose().xyz(),
    ])


class Thrower(MJCFBasedRobot):
  min_target_placement_radius = 0.1
  max_target_placement_radius = 0.8
  min_object_placement_radius = 0.1
  max_object_placement_radius = 0.8

  def __init__(self):
    MJCFBasedRobot.__init__(self, 'thrower.xml', 'body0', action_dim=7, obs_dim=48)

  def robot_specific_reset(self, bullet_client):
    # parts
    self.fingertip = self.parts["fingertip"]
    self.target = self.parts["target"]
    self.object = self.parts["object"]

    # joints
    self.shoulder_pan_joint = self.jdict["shoulder_pan_joint"]
    self.shoulder_lift_joint = self.jdict["shoulder_lift_joint"]
    self.upper_arm_roll_joint = self.jdict["upper_arm_roll_joint"]
    self.elbow_flex_joint = self.jdict["elbow_flex_joint"]
    self.forearm_roll_joint = self.jdict["forearm_roll_joint"]
    self.wrist_flex_joint = self.jdict["wrist_flex_joint"]
    self.wrist_roll_joint = self.jdict["wrist_roll_joint"]

    self._object_hit_ground = False
    self._object_hit_location = None

    # reset position and speed of manipulator
    # TODO: Will this work or do we have to constrain this resetting in some way?
    self.shoulder_pan_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14), 0)
    self.shoulder_lift_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14),
                                                    0)
    self.upper_arm_roll_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14),
                                                     0)
    self.elbow_flex_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14), 0)
    self.forearm_roll_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14), 0)
    self.wrist_flex_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14), 0)
    self.wrist_roll_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14), 0)

    self.zero_offset = np.array([0.45, 0.55, 0])
    self.object_pos = np.concatenate([
        self.np_random.uniform(low=-1, high=1, size=1),
        self.np_random.uniform(low=-1, high=1, size=1),
        self.np_random.uniform(low=-1, high=1, size=1)
    ])

    # make length of vector between min and max_object_placement_radius
    self.object_pos = self.object_pos \
          / np.linalg.norm(self.object_pos) \
          * self.np_random.uniform(low=self.min_object_placement_radius,
                 high=self.max_object_placement_radius, size=1)

    # reset object position
    self.parts["object"].reset_pose(self.object_pos - self.zero_offset, np.array([0, 0, 0, 1]))

    self.target_pos = np.concatenate([
        self.np_random.uniform(low=-1, high=1, size=1),
        self.np_random.uniform(low=-1, high=1, size=1),
        self.np_random.uniform(low=-1, high=1, size=1)
    ])

    # make length of vector between min and max_target_placement_radius
    self.target_pos = self.target_pos \
          / np.linalg.norm(self.target_pos) \
          * self.np_random.uniform(low=self.min_target_placement_radius,
                 high=self.max_target_placement_radius, size=1)

    self.parts["target"].reset_pose(self.target_pos - self.zero_offset, np.array([0, 0, 0, 1]))

  def apply_action(self, a):
    assert (np.isfinite(a).all())
    self.shoulder_pan_joint.set_motor_torque(0.05 * float(np.clip(a[0], -1, +1)))
    self.shoulder_lift_joint.set_motor_torque(0.05 * float(np.clip(a[1], -1, +1)))
    self.upper_arm_roll_joint.set_motor_torque(0.05 * float(np.clip(a[2], -1, +1)))
    self.elbow_flex_joint.set_motor_torque(0.05 * float(np.clip(a[3], -1, +1)))
    self.forearm_roll_joint.set_motor_torque(0.05 * float(np.clip(a[4], -1, +1)))
    self.wrist_flex_joint.set_motor_torque(0.05 * float(np.clip(a[5], -1, +1)))
    self.wrist_roll_joint.set_motor_torque(0.05 * float(np.clip(a[6], -1, +1)))

  def calc_state(self):
    self.to_target_vec = self.target_pos - self.object_pos
    return np.concatenate([
        np.array([j.current_position() for j in self.ordered_joints]).flatten(),  # all positions
        np.array([j.current_relative_position() for j in self.ordered_joints
                 ]).flatten(),  # all speeds
        self.to_target_vec,
        self.fingertip.pose().xyz(),
        self.object.pose().xyz(),
        self.target.pose().xyz(),
    ])



"""A wrapped Quadruped with State-machine based controller."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import enum
import gin
import gym
import numpy as np

NUM_LEGS = 4
ACTION_DIM_COM = 2
ACTION_DIM_TOE = 1
ACTION_DIM_TOTAL = ACTION_DIM_COM + ACTION_DIM_TOE
OBSERVATION_DIM_LEG_ID = NUM_LEGS
OBSERVATION_DIM_TOE_POS = 2 * NUM_LEGS
OBSERVATION_DIM_TOTAL = OBSERVATION_DIM_TOE_POS + OBSERVATION_DIM_LEG_ID


# States of the state machine.
class GaitStateMachine(enum.IntEnum):
  """The state machine for quadruped gait."""
  STEP_LEFT_FRONT_TOE = 0
  STEP_RIGHT_HIND_TOE = 1
  STEP_RIGHT_FRONT_TOE = 2
  STEP_LEFT_HIND_TOE = 3
  TOTAL_GAIT_STATE_NUM = 4


@gin.configurable
class StateMachineBasedWrapperEnv(object):
  """An env using IK to convert toe positions to joint angles.

    The state machine consists of 4 states. During each state, the center of
    mass of the base link is moved first, then one of the legs will take a
    step by following a planned elliptical trajectory. The legs will move in
    the order of front left -> hind right -> front right -> hind left.
    The state transition is determined by elapsed time since last transition.
    Observation (16 dimensions):
    [one hot vector of the state id, local toe positions in x and y direction]
    Action (3 dimensions):
    [target moving distance of the current moving leg in x direction,
     target moving distance of base COM in x direction,
     target moving distance of base COM in y direction]
  """

  def __init__(self,
               gym_env,
               default_local_toe_positions,
               toe_link_indices=(3, 7, 11, 15),
               foot_lift_height=0.15,
               state_duration=2.0,
               action_lower_bound=(-0.0, -0.25, -0.25),
               action_upper_bound=(0.3, 0.25, 0.25),
               state_to_foot_id=(0, 3, 2, 1)):
    """Initialzes the wrapped env.

    Args:
      gym_env: An instance of LocomotionGymEnv.
      default_local_toe_positions: A list of vectors that contains the default
        local position of each toe.
      toe_link_indices: A list of indices to the toe link. Used for calculating
        local toe positions.
      foot_lift_height: Specifies how high the foot lifts during swing stage.
      state_duration: Specifies the duration of each state.
      action_lower_bound: Lower bound for the actions.
      action_upper_bound: Upper bound for the actions.
      state_to_foot_id: Mapping from state machine state to foot id.
    """
    self._gym_env = gym_env

    assert len(action_lower_bound) == ACTION_DIM_TOTAL
    assert len(action_upper_bound) == ACTION_DIM_TOTAL
    self.action_space = gym.spaces.Box(
        np.array(action_lower_bound), np.array(action_upper_bound))
    observation_lower_bound = np.array([-1.0] * OBSERVATION_DIM_TOTAL)
    observation_upper_bound = np.array([1.0] * OBSERVATION_DIM_TOTAL)
    self.observation_space = gym.spaces.Box(observation_lower_bound,
                                            observation_upper_bound)
    self._default_local_toe_positions = default_local_toe_positions
    self._toe_link_indices = toe_link_indices
    self._foot_lift_height = foot_lift_height
    self._state_to_foot_id = state_to_foot_id

    # Use the largest value for bounding the toe movement
    self._toe_move_bound = np.max(
        np.abs(np.concatenate([action_lower_bound, action_upper_bound])))

    # Duration of each state
    self.state_machine_state_duration = [state_duration] * int(
        GaitStateMachine.TOTAL_GAIT_STATE_NUM)

  def __getattr__(self, attr):
    return getattr(self._gym_env, attr)

  def _state_machine_observation(self):
    """Get the current observation: [state_id, local_toe_positions]."""
    observation = []
    # One-hot vector for the current state machine state.
    one_hot_foot_id = np.zeros(GaitStateMachine.TOTAL_GAIT_STATE_NUM)
    one_hot_foot_id[self.current_state_machine_state] = 1
    observation.extend(one_hot_foot_id)

    # Toe positions in X and Y direction in the local space.
    for toe_index in range(NUM_LEGS):
      toe_pos_local = self.current_local_toe_positions[toe_index]
      toe_pos_local_xy = [toe_pos_local[0], toe_pos_local[1]]
      observation.extend(toe_pos_local_xy)

    return observation

  def _get_constant_accel_interpolation(self, interp_ratio):
    """Modify an interpolation between 0 and 1 to have constant acceleration."""
    assert interp_ratio <= 1.0 and interp_ratio >= 0.0
    if interp_ratio < 0.5:
      return 0.5 * (2 * interp_ratio)**2
    else:
      return -0.5 * (2 * interp_ratio - 2)**2 + 1

  def _move_com(self, target_com_movement, time_since_transition,
                state_duration):
    """Get the ik action for moving the COM.

    Args:
      target_com_movement: Target COM movement relative to the previous COM
        position in the x-y plane.
      time_since_transition: Time elapsed since last state machien transition.
      state_duration: Duration of the state machine. The first half will be used
        for moving COM and the second half for moving the swing leg.

    Returns:
      ik_action: ik targets for moving the COM.
    """
    com_movement_ratio = np.clip(
        time_since_transition / ((state_duration - 0.0) / 2.0), 0, 1)
    com_movement_ratio = self._get_constant_accel_interpolation(
        com_movement_ratio)
    current_com_movement = np.array(target_com_movement) * com_movement_ratio

    ik_action = []
    for toe_index in range(NUM_LEGS):
      toe_pos_local = np.copy(self.current_local_toe_positions[toe_index])
      toe_pos_local[2] = self._default_local_toe_positions[toe_index][2]
      toe_pos_local[0] -= current_com_movement[0]
      toe_pos_local[1] -= current_com_movement[1]
      toe_pos_local[0] = np.clip(
          toe_pos_local[0], self._default_local_toe_positions[toe_index][0] -
          self._toe_move_bound,
          self._default_local_toe_positions[toe_index][0] +
          self._toe_move_bound)
      toe_pos_local[1] = np.clip(
          toe_pos_local[1], self._default_local_toe_positions[toe_index][1] -
          self._toe_move_bound,
          self._default_local_toe_positions[toe_index][1] +
          self._toe_move_bound)
      ik_action.extend(toe_pos_local)

    zero_translation = [0, 0, 0]
    identity_rotation = [0, 0, 0, 1]
    ik_action.extend(zero_translation)
    ik_action.extend(identity_rotation)

    return ik_action

  def _move_leg(self, target_toe_movement, time_since_transition,
                state_duration):
    """Get the ik action for moving the swing leg.

    Args:
      target_toe_movement: Target toe movement relative to the default toe
        position in the positive x direction.
      time_since_transition: Time elapsed since last state machien transition.
      state_duration: Duration of the state machine. The first half will be used
        for moving COM and the second half for moving the swing leg.

    Returns:
      ik_action: ik targets for moving the swing leg.
    """

    # The target toe position at the end of the movement.
    target_toe_local_position = np.array(self._default_local_toe_positions[
        self._state_to_foot_id[self.current_state_machine_state]])
    target_toe_local_position[0] += target_toe_movement

    # Toe position at the beginning of the movement.
    initial_toe_local_position = np.array(self.current_local_toe_positions[
        self._state_to_foot_id[self.current_state_machine_state]])

    # Auxiliary variables for computing the interpolation between the current
    # toe position and the target toe position.
    toe_circle_radius = 0.5 * np.linalg.norm(
        np.array(target_toe_local_position) - initial_toe_local_position)
    toe_moving_direction = np.array(
        target_toe_local_position) - initial_toe_local_position
    toe_moving_direction /= np.max([np.linalg.norm(toe_moving_direction), 1e-5])
    toe_traj_scale_ratio = self._foot_lift_height / np.max(
        [toe_circle_radius, 1e-5])

    # Current percentage of time into the state machine.
    toe_movement_ratio = np.clip(
        (time_since_transition - state_duration / 2.0) /
        ((state_duration - 0.0) / 2.0), 0, 1)
    toe_movement_ratio = self._get_constant_accel_interpolation(
        toe_movement_ratio)
    toe_circle_moved_angle = np.pi * toe_movement_ratio

    # Target horizontal movement from the previous local toe position.
    target_toe_horizontal_movement = toe_circle_radius - np.cos(
        toe_circle_moved_angle) * toe_circle_radius
    current_toe_target = np.array([
        target_toe_horizontal_movement * toe_moving_direction[0],
        target_toe_horizontal_movement * toe_moving_direction[1],
        np.sin(toe_circle_moved_angle) * toe_circle_radius *
        toe_traj_scale_ratio
    ]) + initial_toe_local_position

    ik_action = []
    for toe_index in range(NUM_LEGS):
      toe_pos_local = np.copy(self.current_local_toe_positions[toe_index])
      toe_pos_local[2] = self._default_local_toe_positions[toe_index][2]
      if toe_index == self._state_to_foot_id[self.current_state_machine_state]:
        toe_pos_local[0] = current_toe_target[0]
        toe_pos_local[1] = current_toe_target[1]
        toe_pos_local[2] = current_toe_target[2]

      ik_action.extend(toe_pos_local)

    zero_translation = [0, 0, 0]
    identity_rotation = [0, 0, 0, 1]
    ik_action.extend(zero_translation)
    ik_action.extend(identity_rotation)
    return ik_action

  def _update_state_machine_transition(self):
    """Update the state machine state if the duration has been reached."""
    self.current_state_machine_state = (self.current_state_machine_state + 1
                                       ) % GaitStateMachine.TOTAL_GAIT_STATE_NUM
    self.last_state_transition_time = self.robot.GetTimeSinceReset()

  def _update_local_toe_positions(self):
    """Update the local position of the toes."""
    identity_orientation = [0, 0, 0, 1]
    base_position = self.robot.GetBasePosition()
    base_orientation = self.robot.GetBaseOrientation()
    inv_base_position, inv_base_orientation = self.pybullet_client.invertTransform(
        base_position, base_orientation)
    for toe_index in range(NUM_LEGS):
      toe_pose_world = self.pybullet_client.getLinkState(
          self.robot.quadruped, self._toe_link_indices[toe_index])[0]
      toe_pos_local, _ = self.pybullet_client.multiplyTransforms(
          inv_base_position, inv_base_orientation, toe_pose_world,
          identity_orientation)
      self.current_local_toe_positions[toe_index] = np.array(toe_pos_local)

  def step(self, action):
    """Steps the wrapped environment.

    Args:
      action: Numpy array. The input action from an NN agent.

    Returns:
      The tuple containing the modified observation, the reward, the epsiode end
      indicator.

    Raises:
      ValueError if input action is None.

    """
    if action is None:
      raise ValueError("Action cannot be None")
    action = np.clip(action, self.action_space.low, self.action_space.high)
    sum_reward = 0
    step_num = 0
    state_duration = self.state_machine_state_duration[
        self.current_state_machine_state]
    time_since_transition = self.robot.GetTimeSinceReset(
    ) - self.last_state_transition_time

    # Move COM
    while time_since_transition < state_duration / 2.0:
      ik_actions = self._move_com(action[1:3], time_since_transition,
                                  state_duration)
      _, reward, done, _ = self._gym_env.step(ik_actions)
      sum_reward += reward
      step_num += 1
      time_since_transition = self.robot.GetTimeSinceReset(
      ) - self.last_state_transition_time
      if done:
        break
    self._update_local_toe_positions()
    # Move Leg
    while time_since_transition < state_duration:
      ik_actions = self._move_leg(action[0], time_since_transition,
                                  state_duration)
      _, reward, done, _ = self._gym_env.step(ik_actions)
      sum_reward += reward
      step_num += 1
      time_since_transition = self.robot.GetTimeSinceReset(
      ) - self.last_state_transition_time
      if done:
        break
    self._update_local_toe_positions()
    self._update_state_machine_transition()

    state_machine_observation = self._state_machine_observation()

    return state_machine_observation, sum_reward, done, _

  def reset(self):
    """Reset the simulation and state machine states."""
    self.current_state_machine_state = GaitStateMachine.STEP_LEFT_FRONT_TOE
    self.last_state_transition_time = 0
    self.current_local_toe_positions = np.copy(
        self._default_local_toe_positions)

    self._gym_env.reset()

    state_machine_observation = self._state_machine_observation()

    return state_machine_observation



#rudimentary MuJoCo mjcf to ROS URDF converter using the UrdfEditor

from pybullet_utils import bullet_client as bc
import pybullet_data as pd

import pybullet_utils.urdfEditor as ed
import argparse
parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('--mjcf',
                    help='MuJoCo xml file to be converted to URDF',
                    default='mjcf/humanoid.xml')
args = parser.parse_args()

p = bc.BulletClient()
p.setAdditionalSearchPath(pd.getDataPath())
objs = p.loadMJCF(args.mjcf, flags=p.URDF_USE_IMPLICIT_CYLINDER)

for o in objs:
  #print("o=",o, p.getBodyInfo(o), p.getNumJoints(o))
  humanoid = objs[o]
  ed0 = ed.UrdfEditor()
  ed0.initializeFromBulletBody(humanoid, p._client)
  robotName = str(p.getBodyInfo(o)[1], 'utf-8')
  partName = str(p.getBodyInfo(o)[0], 'utf-8')

  print("robotName=", robotName)
  print("partName=", partName)

  saveVisuals = False
  ed0.saveUrdf(robotName + "_" + partName + ".urdf", saveVisuals)



"""State estimator for robot height."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import copy
from typing import Any, Sequence

import gin

from pybullet_envs.minitaur.agents.baseline_controller import state_estimator


@gin.configurable
class COMHeightEstimator(state_estimator.StateEstimatorBase):
  """Estimate the CoM height using base orientation and local toe positions."""

  def __init__(
      self,
      robot: Any,
      com_estimate_leg_indices: Sequence[int] = (0, 1, 2, 3),
      initial_com_height: float = 0.45,
  ):
    """Initializes the class.

    Args:
      robot: A quadruped robot.
      com_estimate_leg_indices: Leg indices used in estimating the CoM height.
      initial_com_height: CoM height used during reset.
    """
    self._robot = robot
    self._com_estimate_leg_indices = com_estimate_leg_indices
    self._initial_com_estimate_leg_indices = copy.copy(com_estimate_leg_indices)
    self._initial_com_height = initial_com_height
    self.reset(0)

  @property
  def estimated_com_height(self):
    return self._com_height

  def reset(self, current_time):
    del current_time
    self._com_height = self._initial_com_height
    self._com_estimate_leg_indices = copy.copy(
        self._initial_com_estimate_leg_indices)

  def update(self, current_time):
    del current_time
    local_toe_poses = self._robot.foot_positions(
        position_in_world_frame=False)
    # We rotate the local toe positions into the world orientation centered
    # at the robot base to estimate the height of the robot.
    world_toe_poses = []
    for toe_p in local_toe_poses:
      world_toe_poses.append(
          self._robot.pybullet_client.multiplyTransforms(
              positionA=(0, 0, 0),
              orientationA=self._robot.base_orientation_quaternion,
              positionB=toe_p,
              orientationB=(0, 0, 0, 1))[0])
    mean_height = 0.0
    num_toe_in_contact = 0
    for leg_id in self._com_estimate_leg_indices:
      mean_height += world_toe_poses[leg_id][2]
      num_toe_in_contact += 1
    mean_height /= num_toe_in_contact
    self._com_height = abs(mean_height)

  @property
  def com_estimate_leg_indices(self):
    return self._com_estimate_leg_indices

  @com_estimate_leg_indices.setter
  def com_estimate_leg_indices(self, leg_indices):
    self._com_estimate_leg_indices = leg_indices



"""Rendering utilities."""

import enum
import math
import os
from typing import Any, Callable, Dict, Iterable, Optional, Text
from absl import logging
import gin
import numpy as np


# These matrices will change by any call to render() and should be saved
# immediately to the local object.
last_used_view_matrix = None
last_used_proj_matrix = None
last_rendered_image_size = None

# Bounds of plane loaded with GEOM_PLANE.
_INFINITY = 1.0e30
# In case of infinite plane, use this scene bounding box.
_DEFAULT_BOUNDING_BOX = ((-15, -15, -10), (15, 15, 0))


def render_image(pybullet_client, width, height, view_matrix, proj_matrix):
  """Renders image as numpy array given view and projection matrices."""
  global last_used_view_matrix, last_used_proj_matrix, last_rendered_image_size
  last_used_view_matrix = view_matrix
  last_used_proj_matrix = proj_matrix

  (_, _, px, _, _) = pybullet_client.getCameraImage(
      width=width,
      height=height,
      renderer=pybullet_client.ER_BULLET_HARDWARE_OPENGL,
      viewMatrix=view_matrix,
      projectionMatrix=proj_matrix)
  rgb_array = np.array(px)
  image = rgb_array[:, :, :3]
  last_rendered_image_size = (image.shape[1], image.shape[0])
  return image


def project_world_to_image(points,
                           view_matrix=None,
                           proj_matrix=None,
                           image_size=None,
                           include_z_coord=False):
  """Projects 3D world-space points to 2D or 3D image-space points.

  If no projection matrices or image_size are given, the last ones are used.
  This means that you can render an image with `render_image` function (or other
  functions that use this one) and then project points to that image immediately
  after that call. Otherwise, you can save `last_used_view_matrix` and
  `last_used_proj_matrix` variables and pass them in later.

  Args:
    points: Sequence of 3D points.
    view_matrix: Optional view matrix (column-major).
    proj_matrix: Optional projection matrix (column-major).
    image_size: Optional image size for resulting coords.
    include_z_coord: Whether to include Z coordinate in the resulting
      projection. Z coordinate goes from 0 (far plane) to 1 (near plane).

  Returns:
    A Numpy array of shape (points_count, 2) or (points_count, 3) when
    `include_z_coord` is True, with dtype of np.float32.
  """
  # Note that these matrices are column-major.
  view_matrix = np.array(
      view_matrix or last_used_view_matrix, dtype=np.float32).reshape((4, 4))
  proj_matrix = np.array(
      proj_matrix or last_used_proj_matrix, dtype=np.float32).reshape((4, 4))
  mvp_matrix = np.matmul(view_matrix, proj_matrix)

  # Add w component equal to 1 (for perspective projection).
  points = np.asarray(points, dtype=np.float32)
  points_4d = np.concatenate(
      [points, np.ones((points.shape[0], 1), dtype=np.float32)], axis=-1)
  points_proj = np.matmul(points_4d, mvp_matrix)

  if include_z_coord:
    # Perspective divide (only keep X, Y, and Z, discard W).
    points_proj_3d = points_proj[:, 0:3] / np.expand_dims(points_proj[:, 3], -1)
    # Shift origin to bottom left and rescale range to [0,1]. This assumes
    # OpenGL projection space.
    points_proj_3d = (points_proj_3d + 1) * 0.5
    # Invert y-axis to have a coordinate with (0,0) on the top left.
    points_proj_3d[:, 1] = 1 - points_proj_3d[:, 1]
    # Scale projection to image size, ignore Z.
    image_size = np.asarray(
        image_size or last_rendered_image_size, dtype=np.float32)
    image_size = np.append(image_size, 1)
    return points_proj_3d * image_size
  else:
    # Perspective divide (only keep X and Y, discard Z and W).
    points_proj_2d = points_proj[:, 0:2] / np.expand_dims(points_proj[:, 3], -1)
    # Shift origin to bottom left and rescale range to [0,1]. This assumes
    # OpenGL projection space.
    points_proj_2d = (points_proj_2d + 1) * 0.5
    # Invert y-axis to have a coordinate with (0,0) on the top left.
    points_proj_2d[:, 1] = 1 - points_proj_2d[:, 1]
    # Scale projection to image size.
    image_size = np.asarray(
        image_size or last_rendered_image_size, dtype=np.float32)
    return points_proj_2d * image_size


def get_scene_bounding_box(pybullet_client, scene=None):
  """Computes scene axis-aligned bounding box.

  Args:
    pybullet_client: PyBullet client.
    scene: Scene instance for filtering the bounding box of camera.

  Returns:
    A tuple of min and max 3D coordinates. Returns (None, None) if the scene
    is empty.
  """
  aabb_min = None
  aabb_max = None

  for i in range(pybullet_client.getNumBodies()):
    body_id = pybullet_client.getBodyUniqueId(i)

    # If a scene has been provided, only count bodes which are in
    # either the ground or obstacle id lists.
    if scene is not None:
      if body_id not in scene.ground_ids and body_id not in scene.obstacle_ids:
        continue

    aabb = pybullet_client.getAABB(body_id)
    if np.any(np.abs(aabb) >= _INFINITY):
      aabb = _DEFAULT_BOUNDING_BOX
    if aabb_min is None:
      aabb_min = aabb[0]
      aabb_max = aabb[1]
    else:
      aabb_min = np.minimum(aabb_min, aabb[0])
      aabb_max = np.maximum(aabb_max, aabb[1])

  return aabb_min, aabb_max


@gin.configurable
def render_topdown(
    pybullet_client,
    result_size=(1280, 720),
    scale_px_per_meter=None,
    camera_height=50,
    ground_height=None,
    low_render_height_from_ground=None,
    high_render_height_from_ground=None,
    scene=None,
    use_y_as_up_axis=False,
    rendered_origin_and_size=None,
):
  """Renders top-down image of the environment.

  Args:
    pybullet_client: PyBullet client.
    result_size: Resulting image size.
    scale_px_per_meter: Resulting image scale in pixels per meter. This
      overrides `result_size`.
    camera_height: Height of the camera above the environment. This is not very
      significant, the lower the height, the larger perspective distortion.
    ground_height: Ground height for following two parameters.
    low_render_height_from_ground: If set, rendering is cut below this distance
      from ground.
    high_render_height_from_ground:  If set, rendering is cut above this
      distance from ground.
    scene: Scene instance for filtering the bounding box of camera.
    use_y_as_up_axis: Whether to consider Y axis as world's "up" instead of Z.
    rendered_origin_and_size: If set, sets rendered origin (bounding box min)
      and size (bounding box size) instead of computing that from the scene.
      Expected format: two tuples of [x, y z] coords containing origin and size
        of rendered bounding box.

  Returns:
    RGB image as 3D numpy array.
  """
  # Get bounds of the current environment.
  if rendered_origin_and_size is not None:
    aabb_min, aabb_size = rendered_origin_and_size
    if len(aabb_min) != 3:
      raise ValueError(
          "Invalid render origin, expected [x, y, z]: {}".format(aabb_min))
    if len(aabb_size) != 3:
      raise ValueError(
          "Invalid render size, expected [x, y, z]: {}".format(aabb_size))
    aabb_max = tuple(m + s for m, s in zip(aabb_min, aabb_size))
  else:
    aabb_min, aabb_max = get_scene_bounding_box(pybullet_client, scene)

  if use_y_as_up_axis:
    width = aabb_max[0] - aabb_min[0]
    height = aabb_max[2] - aabb_min[2]
    z_size = aabb_max[1] - aabb_min[1]
    z_max = aabb_max[1]
  else:
    width = aabb_max[0] - aabb_min[0]
    height = aabb_max[1] - aabb_min[1]
    z_size = aabb_max[2] - aabb_min[2]
    z_max = aabb_max[2]

  if scale_px_per_meter is not None:
    result_size = (int(width * scale_px_per_meter),
                   int(height * scale_px_per_meter))
  else:
    # Adjust scene size to fit inside of the given result size.
    if len(result_size) != 2 or result_size[0] <= 0 or result_size[1] <= 0:
      raise ValueError("Invalid result size: {}".format(result_size))
    img_width, img_height = result_size
    if img_width / width < img_height / height:
      # Width is limiting. Adjust height to match.
      adjusted_height = img_height * width / img_width
      assert adjusted_height >= height, (adjusted_height, height)
      height = adjusted_height
    else:
      # Height is limiting. Adjust width to match.
      adjusted_width = img_width * height / img_height
      assert adjusted_width >= width, (adjusted_width, width)
      width = adjusted_width

  # Compute view and projection matrices.
  if use_y_as_up_axis:
    center_x = (aabb_min[0] + aabb_max[0]) / 2.0
    center_z = (aabb_min[2] + aabb_max[2]) / 2.0
    view_matrix = pybullet_client.computeViewMatrix(
        cameraEyePosition=(center_x, aabb_max[1] + camera_height, center_z),
        cameraTargetPosition=(center_x, aabb_max[1], center_z),
        cameraUpVector=(0, 0, 1))
  else:
    center_x = (aabb_min[0] + aabb_max[0]) / 2.0
    center_y = (aabb_min[1] + aabb_max[1]) / 2.0
    view_matrix = pybullet_client.computeViewMatrix(
        cameraEyePosition=(center_x, center_y, aabb_max[2] + camera_height),
        cameraTargetPosition=(center_x, center_y, aabb_max[2]),
        cameraUpVector=(0, 1, 0))

  near_plane = camera_height
  far_plane = camera_height + z_size
  if ground_height is not None:
    if low_render_height_from_ground is not None:
      far_plane = (
          camera_height + z_max - ground_height - low_render_height_from_ground)
    if high_render_height_from_ground is not None:
      near_plane = (
          camera_height + z_max - ground_height -
          high_render_height_from_ground)
  else:
    if (low_render_height_from_ground is not None or
        high_render_height_from_ground is not None):
      raise ValueError(
          "The `low_render_height_from_ground` or "
          "`high_render_height_from_ground` were specified but no reference "
          "ground height was given.")
  vertical_fov = 2 * math.atan2(height / 2, camera_height)
  proj_matrix = pybullet_client.computeProjectionMatrixFOV(
      fov=math.degrees(vertical_fov),
      aspect=width / height,
      nearVal=near_plane,
      farVal=far_plane)

  # Render and return image.
  return render_image(
      pybullet_client,
      result_size[0],
      result_size[1],
      view_matrix,
      proj_matrix,
  )


